# This file is part of HDL Checker.
#
# Copyright (c) 2015 - 2019 suoto (Andre Souto)
#
# HDL Checker is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HDL Checker is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with HDL Checker.  If not, see <http://www.gnu.org/licenses/>.
"Xilinx xhvdl builder implementation"

import os
import os.path as p
import re
import shutil
import tempfile
from typing import Iterable, Optional

from .base_builder import BaseBuilder

from hdl_checker.diagnostics import BuilderDiag, DiagType
from hdl_checker.parsers.elements.identifier import Identifier
from hdl_checker.path import Path
from hdl_checker.types import BuildFlags, FileType
from hdl_checker.utils import runShellCommand


class XVHDL(BaseBuilder):
    """Builder implementation of the xvhdl compiler"""

    # Implementation of abstract class properties
    builder_name = "xvhdl"
    # TODO: Add xvlog support
    file_types = {FileType.vhdl}

    # XVHDL specific class properties
    _stdout_message_scanner = re.compile(
        r"^(?P<severity>[EW])\w+:\s*"
        r"\[(?P<error_code>[^\]]+)\]\s*"
        r"(?P<error_message>[^\[]+)\s*"
        r"("
        r"\[(?P<filename>[^:]+):"
        r"(?P<line_number>\d+)\]"
        r")?",
        flags=re.I,
    ).scanner

    _iter_rebuild_units = re.compile(
        r"ERROR:\s*\[[^\]]*\]\s*"
        r"'?.*/(?P<library_name>\w+)/(?P<unit_name>\w+)\.vdb'?"
        r"\s+needs to be re-saved.*",
        flags=re.I,
    ).finditer

    def _shouldIgnoreLine(self, line):
        # type: (str) -> bool
        if "ignored due to previous errors" in line:
            return True

        # Ignore messages like
        # ERROR: [VRFC 10-3032] 'library.package' failed to restore
        # This message doesn't come alone, we should be getting other (more
        # usefull) info anyway
        if "[VRFC 10-3032]" in line:
            return True

        return not (line.startswith("ERROR") or line.startswith("WARNING"))

    def __init__(self, *args, **kwargs):
        # type: (...) -> None
        self._version = ""
        super(XVHDL, self).__init__(*args, **kwargs)
        self._xvhdlini = p.join(self._work_folder, ".xvhdl.init")
        # Create the ini file
        open(self._xvhdlini, "w").close()

    def _makeRecords(self, line):
        # type: (str) -> Iterable[BuilderDiag]
        scan = self._stdout_message_scanner(line)

        match = scan.match()
        if not match:
            return

        info = match.groupdict()

        diag = BuilderDiag(
            builder_name=self.builder_name,
            text=info["error_message"].strip(),
            line_number=info["line_number"],
            filename=Path(info["filename"]),
            error_code=info["error_code"],
        )

        if info.get("severity", None) in ("W", "e"):
            diag.severity = DiagType.WARNING
        elif info.get("severity", None) in ("E", "e"):
            diag.severity = DiagType.ERROR

        yield diag

    def _parseBuiltinLibraries(self):
        "(Not used by XVHDL)"
        return (
            Identifier(x, case_sensitive=False)
            for x in (
                "ieee",
                "std",
                "unisim",
                "xilinxcorelib",
                "synplify",
                "synopsis",
                "maxii",
                "family_support",
            )
        )

    def _checkEnvironment(self):
        stdout = runShellCommand(
            ["xvhdl", "--nolog", "--version"], cwd=self._work_folder
        )
        self._version = re.findall(r"^Vivado Simulator\s+([\d\.]+)", stdout[0])[0]
        self._logger.info(
            "xvhdl version string: '%s'. " "Version number is '%s'",
            stdout[:-1],
            self._version,
        )

    @staticmethod
    def isAvailable():
        try:
            temp_dir = tempfile.mkdtemp()
            runShellCommand(["xvhdl", "--nolog", "--version"], cwd=temp_dir)
            return True
        except OSError:
            return False
        finally:
            shutil.rmtree(temp_dir)

    def _createLibrary(self, library):
        # type: (Identifier) -> None
        if not p.exists(self._work_folder):
            os.makedirs(self._work_folder)

        with open(self._xvhdlini, mode="w") as fd:
            content = "\n".join(
                [
                    "%s=%s" % (x, p.join(self._work_folder, x.name))
                    for x in self._added_libraries
                ]
            )
            fd.write(content)

    def _buildSource(self, path, library, flags=None):
        # type: (Path, Identifier, Optional[BuildFlags]) -> Iterable[str]
        cmd = [
            "xvhdl",
            "--nolog",
            "--verbose",
            "0",
            "--initfile",
            self._xvhdlini,
            "--work",
            library.name,
        ]
        cmd += [str(x) for x in (flags or [])]
        cmd += [path.name]
        return runShellCommand(cmd, cwd=self._work_folder)

    def _searchForRebuilds(self, line):
        rebuilds = []

        for match in self._iter_rebuild_units(line):
            mdict = match.groupdict()
            # When compilers reports units out of date, they do this
            # by either
            #  1. Giving the path to the file that needs to be rebuilt
            #     when sources are from different libraries
            #  2. Reporting which design unit has been affected by a
            #     given change.
            if "rebuild_path" in mdict and mdict["rebuild_path"] is not None:
                rebuilds.append(mdict)
            else:
                rebuilds.append(
                    {"library_name": "work", "unit_name": mdict["unit_name"]}
                )

        return rebuilds



# This file is part of HDL Checker.
#
# Copyright (c) 2015 - 2019 suoto (Andre Souto)
#
# HDL Checker is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HDL Checker is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with HDL Checker.  If not, see <http://www.gnu.org/licenses/>.
"Serialization specifics"

import json
import logging
from pprint import pformat

from hdl_checker import path, types
from hdl_checker.builders.fallback import Fallback
from hdl_checker.builders.ghdl import GHDL
from hdl_checker.builders.msim import MSim
from hdl_checker.builders.xvhdl import XVHDL
from hdl_checker.database import Database
from hdl_checker.parsers.elements.dependency_spec import DependencySpec
from hdl_checker.parsers.elements.design_unit import VerilogDesignUnit, VhdlDesignUnit
from hdl_checker.parsers.elements.identifier import (
    Identifier,
    VerilogIdentifier,
    VhdlIdentifier,
)
from hdl_checker.parsers.verilog_parser import VerilogParser
from hdl_checker.parsers.vhdl_parser import VhdlParser

_logger = logging.getLogger(__name__)

# Maps class names added by the decoder to the actual class on Python side to
# recreate an object
CLASS_MAP = {
    "Database": Database,
    "DependencySpec": DependencySpec,
    "Fallback": Fallback,
    "FileType": types.FileType,
    "GHDL": GHDL,
    "Identifier": Identifier,
    "MSim": MSim,
    "Path": path.Path,
    "TemporaryPath": path.TemporaryPath,
    "VerilogIdentifier": VerilogIdentifier,
    "VerilogParser": VerilogParser,
    "VhdlDesignUnit": VhdlDesignUnit,
    "VerilogDesignUnit": VerilogDesignUnit,
    "VhdlIdentifier": VhdlIdentifier,
    "VhdlParser": VhdlParser,
    "XVHDL": XVHDL,
}


class StateEncoder(json.JSONEncoder):
    """
    Custom encoder that handles hdl_checker classes
    """

    def default(self, o):  # pylint: disable=method-hidden
        if hasattr(o, "__jsonEncode__"):
            dct = o.__jsonEncode__()
            # Set a __class__ entry into the dict so we can use it to get from
            # CLASS_MAP when recreating the object
            prev = dct.get("__class__", None)
            if prev is not None and o.__class__.__name__ != prev:
                _logger.warning(
                    "Class has been set to %s, will NOT overwrite it with %s!",
                    prev,
                    o.__class__.__name__,
                )
            else:
                dct["__class__"] = o.__class__.__name__
            #  _logger.debug("object: %s, result:\n%s", repr(o), pformat(dct))
            return dct
        # Let the base class default method raise the TypeError
        try:
            return json.JSONEncoder.default(self, o)
        except:
            _logger.fatal("object: %s, type=%s", o, type(o))
            raise


def jsonObjectHook(dict_):
    """
    json hook for decoding entries added the StateEncoder back to Python
    objects
    """
    if "__class__" not in dict_:
        return dict_

    cls_name = dict_["__class__"]
    cls = CLASS_MAP.get(cls_name, None)
    assert cls is not None, "We should handle {}".format(cls_name)

    try:
        obj = cls.__jsonDecode__(dict_)
    except:  # pragma: no cover
        _logger.error("Something went wrong, cls_name: %s => %s", cls_name, cls)
        raise
    return obj



# This file is part of HDL Checker.
#
# Copyright (c) 2015 - 2019 suoto (Andre Souto)
#
# HDL Checker is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HDL Checker is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with HDL Checker.  If not, see <http://www.gnu.org/licenses/>.

# pylint: disable=function-redefined, missing-docstring, protected-access

import json
import logging
import os
import os.path as p
import subprocess as subp
import tempfile
import time
from multiprocessing import Event, Process, Queue
from threading import Thread

import mock
import requests
import six
from pyls import uris  # type: ignore
from pyls.python_ls import PythonLanguageServer, start_io_lang_server  # type: ignore

from nose2.tools import such  # type: ignore

from hdl_checker.tests import disableVunit, getTestTempPath

import hdl_checker.lsp
from hdl_checker.utils import isProcessRunning, ON_WINDOWS, terminateProcess

_logger = logging.getLogger(__name__)

TEST_TEMP_PATH = getTestTempPath(__name__)
TEST_LOG_PATH = p.join(os.environ["TOX_ENV_DIR"], "log")
SERVER_LOG_LEVEL = os.environ.get("SERVER_LOG_LEVEL", "WARNING")

HDL_CHECKER_BASE_PATH = p.abspath(p.join(p.dirname(__file__), "..", ".."))

JSONRPC_VERSION = "2.0"
CALL_TIMEOUT = 5


def _path(*args):
    # type: (str) -> str
    "Helper to reduce foorprint of p.join(TEST_TEMP_PATH, *args)"
    return p.join(TEST_TEMP_PATH, *args)


def doNothing(queue):
    _logger.debug("I'm ready")
    queue.get()
    _logger.debug("Ok, done")


def _getUnusedLocalhostPort():
    """
    These were "Borrowed" from YCM.
    See https://github.com/Valloric/YouCompleteMe
    """
    import socket

    sock = socket.socket()
    # This tells the OS to give us any free port in the range [1024 - 65535]
    sock.bind(("", 0))
    port = sock.getsockname()[1]
    sock.close()
    return port


def _startClient(client):
    client.start()


class _ClientServer(
    object
):  # pylint: disable=useless-object-inheritance,too-few-public-methods
    """ A class to setup a client/server pair """

    def __init__(self):
        # Client to Server pipe
        csr, csw = os.pipe()
        # Server to client pipe
        scr, scw = os.pipe()

        self.server_thread = Thread(
            target=start_io_lang_server,
            args=(
                os.fdopen(csr, "rb"),
                os.fdopen(scw, "wb"),
                False,
                hdl_checker.lsp.HdlCheckerLanguageServer,
            ),
        )

        self.server_thread.daemon = True
        self.server_thread.start()

        # Object being tested is the server thread. Avoid both objects
        # competing for the same cache by using the raw Python language server
        self.client = PythonLanguageServer(
            os.fdopen(scr, "rb"), os.fdopen(csw, "wb"), start_io_lang_server
        )

        self.client_thread = Thread(target=_startClient, args=[self.client])
        self.client_thread.daemon = True
        self.client_thread.start()


such.unittest.TestCase.maxDiff = None

with such.A("hdl_checker server") as it:

    _SERVER_BASE_CMD = [
        "coverage",
        "run",
        p.join(HDL_CHECKER_BASE_PATH, "hdl_checker", "server.py"),
        "--log-level",
        SERVER_LOG_LEVEL,
        "--stdout",
        p.join(TEST_LOG_PATH, "hdl_checker-stdout.log"),
        "--stderr",
        p.join(TEST_LOG_PATH, "hdl_checker-stderr.log"),
        "--log-stream",
        p.join(TEST_LOG_PATH, "tests.log"),
    ]

    with it.having("http server"):

        def startCodeCheckerServer():
            it._host = "127.0.0.1"
            it._port = str(_getUnusedLocalhostPort())
            it._url = "http://{0}:{1}".format(it._host, it._port)

            cmd = list(_SERVER_BASE_CMD) + ["--host", it._host, "--port", str(it._port)]

            _logger.info("Starting hdl_checker server with '%s'", " ".join(cmd))

            stdout_r, stdout_w = os.pipe()
            stderr_r, stderr_w = os.pipe()

            it.stdout = os.fdopen(stdout_r, "rb")
            it.stderr = os.fdopen(stderr_r, "rb")

            it._server = subp.Popen(
                cmd,
                env=os.environ.copy(),
                stdout=os.fdopen(stdout_w, "wb"),
                stderr=os.fdopen(stderr_w, "wb"),
            )
            waitForServer()

        def startServerAttachedToPid(pid):
            it._url = "http://{0}:{1}".format(it._host, it._port)

            cmd = list(_SERVER_BASE_CMD) + [
                "--host",
                it._host,
                "--port",
                str(it._port),
                "--attach-to-pid",
                str(pid),
            ]

            _logger.info("Starting hdl_checker server with '%s'", " ".join(cmd))

            it._server = subp.Popen(cmd, env=os.environ.copy())
            waitForServer()

        def waitForServer():
            event = Event()

            def wait():
                # Wait until the server is up and replying
                start = time.time()
                while not event.is_set():
                    try:
                        reply = requests.post(it._url + "/get_diagnose_info")
                        if reply.ok:
                            _logger.info(
                                "Server replied OK after %.1fs", time.time() - start
                            )
                            event.set()
                            return
                    except requests.ConnectionError:
                        pass
                    except:
                        _logger.exception(
                            "Exception while waiting for server to respond"
                        )
                        raise

                    time.sleep(0.5)

                _logger.info("Exiting wait thread")

            Thread(target=wait).start()
            # Wait 10s for the server to start responding
            event.wait(timeout=10)

            if event.is_set():
                return

            # Set the event from here to force the wait thread to exit
            event.set()

            _logger.error("Server is not replying")

            it._server.terminate()
            terminateProcess(it._server.pid)

            _logger.error("stderr: %s", it.stderr.read())

            it.fail("Server is not responding")

        def waitUntilBuildFinishes(data):
            _logger.info("Waiting for 30s until build is finished")
            for i in range(30):
                #  time.sleep(1)
                _logger.info("Elapsed %ds", i)
                _ = requests.post(it._url + "/get_messages_by_path", data)
                ui_messages = requests.post(it._url + "/get_ui_messages", data)
                _logger.debug("==> %s", ui_messages.json)
                if ui_messages.json["ui_messages"] == []:
                    _logger.info("Ok, done")
                    return

            assert False, "Server is still building after 30s"

        @it.has_teardown
        def teardown():
            it._server.terminate()
            terminateProcess(it._server.pid)

        @it.should("start and respond a request")
        @disableVunit
        def test():
            startCodeCheckerServer()
            some_project = _path("some_project")
            open(some_project, "w").write("")
            # Ensure the server is active
            reply = requests.post(
                it._url + "/get_diagnose_info", data={"project_file": some_project}
            )
            it.assertTrue(reply.ok, "Reply was not OK: {}".format(reply))

        @it.should("shutdown the server when requested")  # type: ignore
        @disableVunit
        def test():
            # Send a request to the shutdown addr
            with it.assertRaises(requests.ConnectionError):
                reply = requests.post(it._url + "/shutdown")
                it.assertFalse(reply.ok)

            it._server.terminate()
            terminateProcess(it._server.pid)

        @it.should(  # type: ignore
            "terminate when the parent PID is not running anymore"
        )
        def test():

            queue = Queue()

            proc = Process(target=doNothing, args=(queue,))
            proc.start()

            _logger.info("Started dummy process with PID %d", proc.pid)
            startServerAttachedToPid(proc.pid)
            time.sleep(3)
            _logger.info("Allowing the dummy process to finish")
            queue.put(1)
            proc.join()

            if isProcessRunning(proc.pid):
                _logger.warning("Dummy process %d was still running", proc.pid)
                proc.terminate()
                time.sleep(1)
                it.assertFalse(
                    isProcessRunning(proc.pid),
                    "Process %d is still running after terminating " "it!" % proc.pid,
                )

            time.sleep(1)
            _logger.info("Server should have died by now")

            with it.assertRaises(requests.ConnectionError):
                requests.post(it._url + "/get_diagnose_info")

    with it.having("LSP server"):

        @it.should("initialize with no project file")  # type: ignore
        @disableVunit
        def test():
            client_server = _ClientServer()
            response = client_server.client._endpoint.request(
                "initialize",
                {
                    "rootPath": uris.from_fs_path(TEST_TEMP_PATH),
                    "initializationOptions": {},
                },
            ).result(timeout=CALL_TIMEOUT)

            _logger.debug("Response: %s", response)
            it.assertEqual(response, {"capabilities": {"textDocumentSync": 1}})

        @it.should("show message with reason for failing to start")  # type: ignore
        @disableVunit
        def test():
            def _start_io_lang_server(*_):  # pylint: disable=invalid-name
                assert False, "Expected fail to trigger the test"

            args = type(
                "args",
                (object,),
                {
                    "lsp": True,
                    "log_level": SERVER_LOG_LEVEL,
                    "stderr": p.join(TEST_LOG_PATH, "hdl_checker-stderr.log"),
                    "log_stream": p.join(TEST_LOG_PATH, "tests.log"),
                    "color": None,
                    "attach_to_pid": None,
                },
            )

            # Python 2 won't allow to mock sys.stdout.write directly
            import sys

            stdout = mock.MagicMock(spec=sys.stdout)
            stdout.write = mock.MagicMock(spec=sys.stdout.write)

            with mock.patch("hdl_checker.server.start_io_lang_server", _start_io_lang_server):
                with mock.patch("hdl_checker.server.sys.stdout", stdout):
                    with it.assertRaises(AssertionError):
                        hdl_checker.server.run(args)

            assertion_msg = "'Expected fail to trigger the test'"

            # Don't know why Python 2 adds an extra comma to the msg
            if six.PY2:
                assertion_msg += ","

            # Build up the expected response
            body = json.dumps(
                {
                    "method": "window/showMessage",
                    "jsonrpc": JSONRPC_VERSION,
                    "params": {
                        "message": "Unable to start HDL Checker LSP server: "
                        "'AssertionError(" + assertion_msg + ")'! "
                        "Check " + p.abspath(args.stderr) + " for more info",
                        "type": 1,
                    },
                }
            )

            response = (
                "Content-Length: {}\r\n"
                "Content-Type: application/vscode-jsonrpc; charset=utf8\r\n\r\n"
                "{}".format(len(body), body)
            )

            stdout.write.assert_called_once_with(response)

    with it.having("LSP server executable"):

        def assertCommandPrints(cmd, stdout, **kwargs):
            _logger.debug("Running command: %s", cmd)
            output = subp.check_output(cmd, **kwargs).decode().strip()
            it.assertEqual(output, stdout)

        @it.should("report version correctly")  # type: ignore
        def test():
            assertCommandPrints(["hdl_checker", "--version"], hdl_checker.__version__)

        def startServerWrapper(cmd):
            log_file = tempfile.mktemp()

            actual_cmd = cmd + ["--nocolor", "--log-stream", log_file]

            _logger.info("Actual command: %s", actual_cmd)

            server = subp.Popen(
                actual_cmd, stdin=subp.PIPE, stdout=subp.PIPE, stderr=subp.PIPE
            )

            # Close stdin so the server exits
            stdout, stderr = server.communicate("")

            it.assertEqual(
                stdout, b"", "stdout should be empty but got\n{}".format(stdout)
            )

            it.assertEqual(
                stderr, b"", "stderr should be empty but got\n{}".format(stdout)
            )

            # On Windows the Popen PID and the *actual* PID don't always match
            # for some reason. Since we're not testing this, just skip the
            # first line
            log_content = open(log_file, "rb").read().decode().split("\n")

            expected = [
                "Starting server. Our PID is {}, no parent PID to attach to. "
                "Version string for hdl_checker is '{}'".format(
                    server.pid, hdl_checker.__version__
                ),
                "Starting HdlCheckerLanguageServer IO language server",
            ]

            _logger.info("Log content: %s", log_content)

            if ON_WINDOWS:
                log_content = log_content[1:]
                expected = expected[1:]

            for line in expected:
                it.assertIn(line, "\n".join(log_content))

            #  it.assertEqual(log_content[: len(expected)], expected)
            #  it.fail("stop")

            os.remove(log_file)

        @it.should(  # type: ignore
            "start server given the --lsp flag and setting stderr"
        )
        def test():
            startServerWrapper(
                [
                    "hdl_checker",
                    "--lsp",
                    "--stderr",
                    p.join(TEST_LOG_PATH, "hdl_checker_stderr.log"),
                ]
            )

        @it.should("start server given the --lsp flag")  # type: ignore
        def test():
            startServerWrapper(["hdl_checker", "--lsp"])


it.createTests(globals())



# This file is part of HDL Checker.
#
# Copyright (c) 2015 - 2019 suoto (Andre Souto)
#
# HDL Checker is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HDL Checker is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with HDL Checker.  If not, see <http://www.gnu.org/licenses/>.
"Spec for a parsed dependency"

import logging
from typing import Optional

from .identifier import Identifier
from .parsed_element import LocationList, ParsedElement

from hdl_checker import types as t  # pylint: disable=unused-import
from hdl_checker.path import Path

_logger = logging.getLogger(__name__)


class DependencySpec(ParsedElement):
    "Placeholder for a source dependency"

    def __init__(self, owner, name, library=None, locations=None):
        # type: (Path, Identifier, Optional[Identifier], Optional[LocationList]) -> None
        assert isinstance(name, Identifier), "Incorrect arg: {}".format(name)
        assert library is None or isinstance(
            library, Identifier
        ), "Incorrect arg: {}".format(library)

        self._owner = owner
        self._library = library
        self._name = name
        assert self._name.name != "identifier", "wtf"
        super(DependencySpec, self).__init__(locations)

    @property
    def owner(self):
        return self._owner

    @property
    def name(self):
        return self._name

    @property
    def library(self):
        return self._library

    @property
    def case_sensitive(self):  # type: () -> bool
        ext = self.owner.split(".")[-1].lower()
        return ext not in t.FileType.vhdl.value

    @property
    def __hash_key__(self):
        return (
            self.owner,
            self.library,
            self.name,
            super(DependencySpec, self).__hash_key__,
        )

    def __jsonEncode__(self):
        #  def __init__(self, owner, name, library=None, locations=None):
        return {
            "owner": self.owner,
            "name": self.name,
            "library": self.library,
            "locations": tuple(self.locations),
        }

    @classmethod
    def __jsonDecode__(cls, state):
        """Returns an object of cls based on a given state"""
        return cls(
            library=state.pop("library"),
            name=state.pop("name"),
            owner=state.pop("owner"),
            locations=state.pop("locations"),
        )
        #  return obj

    def __repr__(self):
        return "{}(name='{}', library='{}', owner={}, locations={})".format(
            self.__class__.__name__,
            repr(self.name),
            repr(self.library),
            repr(self.owner),
            repr(self.locations),
        )



# This file is part of HDL Checker.
#
# Copyright (c) 2015 - 2019 suoto (Andre Souto)
#
# HDL Checker is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HDL Checker is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with HDL Checker.  If not, see <http://www.gnu.org/licenses/>.
"Common type definitions for type hinting"
from collections import namedtuple
from enum import Enum
from typing import Tuple, Union

from hdl_checker.path import Path

BuildFlags = Tuple[str, ...]
LibraryAndUnit = namedtuple("LibraryAndUnit", ["library", "unit"])

RebuildUnit = namedtuple("RebuildUnit", ["name", "type_"])
RebuildLibraryUnit = namedtuple("RebuildLibraryUnit", ["name", "library"])
RebuildPath = namedtuple("RebuildPath", ["path"])

RebuildInfo = Union[RebuildUnit, RebuildLibraryUnit, RebuildPath]


class UnknownTypeExtension(Exception):
    """
    Exception thrown when trying to get the file type of an unknown extension.
    Known extensions are one of '.vhd', '.vhdl', '.v', '.vh', '.sv', '.svh'
    """

    def __init__(self, path):
        super(UnknownTypeExtension, self).__init__()
        self._path = path

    def __str__(self):
        return "Couldn't determine file type for path '%s'" % self._path


class FileType(Enum):
    "RTL file types"
    vhdl = "vhdl"
    verilog = "verilog"
    systemverilog = "systemverilog"

    @staticmethod
    def fromPath(path):
        # type: (Path) -> FileType
        "Extracts FileType from the given path's extension"
        ext = path.name.split(".")[-1].lower()
        if ext in ("vhd", "vhdl"):
            return FileType.vhdl
        if ext in ("v", "vh"):
            return FileType.verilog
        if ext in ("sv", "svh"):
            return FileType.systemverilog
        raise UnknownTypeExtension(path)

    def __jsonEncode__(self):
        """
        Gets a dict that describes the current state of this object
        """
        return {"value": self.name}

    @classmethod
    def __jsonDecode__(cls, state):
        """Returns an object of cls based on a given state"""
        return cls(state["value"])


class BuildFlagScope(Enum):
    """
    Scopes of a given set of flags. Values of the items control the actual
    fields extracted from the JSON config
    """

    single = "single"
    dependencies = "dependencies"
    all = "global"



#!/usr/bin/env python
# This file is part of HDL Checker.
#
# Copyright (c) 2015 - 2019 suoto (Andre Souto)
#
# HDL Checker is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HDL Checker is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with HDL Checker.  If not, see <http://www.gnu.org/licenses/>.
"HDL Checker standalone stuff"

from __future__ import print_function

import argparse
import logging
import os
import os.path as p
import sys
import time

from prettytable import PrettyTable

import hdl_checker
from hdl_checker.utils import setupLogging

try:
    import cProfile as profile
except ImportError:
    import profile  # type: ignore

try:
    import argcomplete  # type: ignore
    _HAS_ARGCOMPLETE = True
except ImportError: # pragma: no cover
    _HAS_ARGCOMPLETE = False


_logger = logging.getLogger(__name__)

def _fileExtentensionCompleter(extension): # pragma: no cover
    "Tab completion for 'extension'"
    def _completer(**kwargs): # pylint: disable=missing-docstring
        prefix = kwargs['prefix']
        if prefix == '':
            prefix = os.curdir

        result = []
        for line in os.listdir(prefix):
            if line.lower().endswith('.' + extension):
                result.append(line)
            elif p.isdir(line):
                result.append("./" + line)

        return result
    return _completer

def parseArguments():
    "Argument parser for standalone hdl_checker"

    if ('--version' in sys.argv[1:]) or ('-V' in sys.argv[1:]):  # pragma: no cover
        print(hdl_checker.__version__)
        sys.exit(0)

    parser = argparse.ArgumentParser()

    # Options
    parser.add_argument('--version', action='store_true',
                        help="Shows hdl_checker version and exit")

    parser.add_argument('--verbose', '-v', action='append_const', const=1,
                        help="""Increases verbose level. Use multiple times to
                                increase more""")

    parser.add_argument('--clean', '-c', action='store_true',
                        help="Cleans the project before building")

    parser.add_argument('--sources', '-s', action='append', nargs='*', default=[],
                        help="""Source(s) file(s) to build individually""") \
                            .completer = _fileExtentensionCompleter('vhd')

    parser.add_argument('--debug-print-sources', action='store_true')
    parser.add_argument('--debug-parse-source-file', action='store_true')
    parser.add_argument('--debug-run-static-check', action='store_true')
    parser.add_argument('--debug-profiling', action='store', nargs='?',
                        metavar='OUTPUT_FILENAME', const='hdl_checker.pstats')

    # Mandatory arguments
    parser.add_argument('project_file', action='store', nargs=1,
                        help="""Configuration file that defines what should be
                        built (lists sources, libraries, build flags and so on""")


    if _HAS_ARGCOMPLETE: # pragma: no cover
        argcomplete.autocomplete(parser)

    args = parser.parse_args()

    # PYTHON_ARGCOMPLETE_OK

    args.project_file = args.project_file[0]

    args.log_level = logging.FATAL
    if args.verbose:
        if len(args.verbose) == 1:
            args.log_level = logging.WARNING
        elif len(args.verbose) == 2:
            args.log_level = logging.INFO
        else:
            args.log_level = logging.DEBUG

    # Planify source list if supplied
    if args.sources:
        args.sources = [source for sublist in args.sources for source in sublist]

    return args

class StandaloneProjectBuilder(hdl_checker.hdl_checker_base.HdlCodeCheckerBase):
    """Implementation of standalone hdl_checker.hdl_checker_base.HdlCodeCheckerBase
    to run via shell"""
    _ui_logger = logging.getLogger('UI')
    def _handleUiInfo(self, message):  # pragma: no cover
        self._ui_logger.info(message)

    def _handleUiWarning(self, message):  # pragma: no cover
        self._ui_logger.warning(message)

    def _handleUiError(self, message):  # pragma: no cover
        self._ui_logger.error(message)

def runStandaloneSourceFileParse(fname):
    """Standalone parser run"""
    from hdl_checker.parsers import VhdlParser, VerilogParser

    extension = fname.lower().split('.')[-1]
    cls = VhdlParser if extension in ('vhd', 'vhdl') else VerilogParser

    source = cls(fname)

    print("Source: %s" % source)

    design_units = source.getDesignUnits()
    if design_units: # pragma: no cover
        print(" - Design_units:")
        for unit in design_units:
            print(" -- %s" % str(unit))
    dependencies = source.getDependencies()
    if dependencies: # pragma: no cover
        print(" - Dependencies:")
        for dependency in dependencies:
            print(" -- %s.%s" % (dependency.library, dependency.name))

def runStandaloneStaticCheck(fname):
    """Standalone source_file.VhdlParser run"""
    from hdl_checker.static_check import getStaticMessages

    lines = [x.decode(errors='ignore') for x in open(fname, mode='rb').readlines()]
    for record in getStaticMessages(lines):
        print(record)

def printSourceDiags(project, source):
    "Print diagnostics for the given source"
    start = time.time()
    records = project.getMessagesByPath(source)
    end = time.time()
    _logger.info("Building source '%s' took %.4fs", source, (end - start))
    for record in records:
        if record.filename is not None:
            message = [record.filename]
        else:
            message = [source]

        location = []
        if record.line_number is not None:
            location += ["line %s" % record.line_number]

        if record.column_number is not None:
            location += ["column %s" % record.column_number]

        if location:
            message += ["(%s)" % ', '.join(location)]

        if record.error_code is None:
            message += ["(%s):" % record.severity]
        else:
            message += ["(%s-%s):" % (record.severity,
                                      record.error_code)]

        message += [record.text]

        print(' '.join(message))


def runner(args):
    "Main runner command processing"

    _logger.info("Creating project object")

    project = StandaloneProjectBuilder(args.project_file)

    if args.clean:
        _logger.info("Cleaning up")
        project.clean()

    if args.debug_print_sources:
        sources = PrettyTable(['Filename', 'Library', 'Flags'])
        sources.align['Filename'] = 'l'
        sources.sortby = 'Library'
        for source in project.getSources():
            sources.add_row([source.filename, source.library, " ".join(source.flags)])
        print(sources)

    for source in args.sources:
        printSourceDiags(project, source)

    if args.debug_parse_source_file:
        for source in args.sources:
            runStandaloneSourceFileParse(source)

    if args.debug_run_static_check:
        for source in args.sources:
            runStandaloneStaticCheck(source)

def main():
    "Main hook for standalone usage"
    start = time.time()
    runner_args = parseArguments()
    setupLogging(sys.stdout, runner_args.log_level)
    logging.root.setLevel(runner_args.log_level)
    #  logging.getLogger('hdl_checker.source_file').setLevel(logging.WARNING)
    logging.getLogger('hdl_checker.config_parser').setLevel(logging.WARNING)
    #  logging.getLogger('hdl_checker.builders').setLevel(logging.INFO)
    logging.getLogger('vunit.project').setLevel(logging.ERROR)

    # Running hdl_checker with threads has two major drawbacks:
    # 1) Makes interrupting it impossible currently because each source
    #    file is parsed on is own thread. Since there can be lots of
    #    sources, interrupting a single thread is not enough. This is
    #    discussed at https://github.com/suoto/hdl_checker/issues/19
    # 2) When profiling, the result expected is of the inner hdl_checker calls
    #    and with threads we have no info. This is discussed at
    #    https://github.com/suoto/hdl_checker/issues/16
    # poor results (see suoto/hdl_checker/issues/16).
    # To circumvent this we disable using threads at all when running
    # via standalone (it's ugly, I know)
    # pylint: disable=protected-access
    StandaloneProjectBuilder._USE_THREADS = False
    # pylint: enable=protected-access

    if runner_args.debug_profiling:
        profile.runctx(
            'runner(runner_args)',
            globals=globals(),
            locals={'runner_args' : runner_args},
            filename=runner_args.debug_profiling, sort=-1)
    else:
        runner(runner_args)
    end = time.time()
    _logger.info("Process took %.2fs", (end - start))

if __name__ == '__main__':
    sys.exit(main())



# This file is part of HDL Checker.
#
# Copyright (c) 2015 - 2019 suoto (Andre Souto)
#
# HDL Checker is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HDL Checker is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with HDL Checker.  If not, see <http://www.gnu.org/licenses/>.
"VHDL static checking to find unused signals, ports and constants."

import logging
import re
from typing import Any, Tuple

#  from hdl_checker.path import Path
from hdl_checker.diagnostics import (
    DiagType,
    LibraryShouldBeOmited,
    ObjectIsNeverUsed,
    StaticCheckerDiag,
)

_logger = logging.getLogger(__name__)

_GET_SCOPE = re.compile(
    "|".join(
        [
            r"^\s*entity\s+(?P<entity_name>\w+)\s+is\b",
            r"^\s*architecture\s+(?P<architecture_name>\w+)\s+of\s+(?P<arch_entity>\w+)",
            r"^\s*package\s+(?P<package_name>\w+)\s+is\b",
            r"^\s*package\s+body\s+(?P<package_body_name>\w+)\s+is\b",
        ]
    ),
    flags=re.I,
).finditer

_NO_SCOPE_OBJECTS = re.compile(
    "|".join(
        [
            r"^\s*library\s+(?P<library>[\w\s,]+)",
            r"^\s*attribute\s+(?P<attribute>[\w\s,]+)\s*:",
        ]
    ),
    flags=re.I,
)

_ENTITY_OBJECTS = re.compile(
    "|".join(
        [
            r"^\s*(?P<port>[\w\s,]+)\s*:\s*(in|out|inout|buffer|linkage)\s+\w+",
            r"^\s*(?P<generic>[\w\s,]+)\s*:\s*\w+",
        ]
    ),
    flags=re.I,
).finditer

_ARCH_OBJECTS = re.compile(
    "|".join(
        [
            r"^\s*constant\s+(?P<constant>[\w\s,]+)\s*:",
            r"^\s*signal\s+(?P<signal>[\w,\s]+)\s*:",
            r"^\s*type\s+(?P<type>\w+)\s*:",
            r"^\s*shared\s+variable\s+(?P<shared_variable>[\w\s,]+)\s*:",
        ]
    ),
    flags=re.I,
).finditer

_SHOULD_END_SCAN = re.compile(
    "|".join(
        [
            r"\bgeneric\s+map",
            r"\bport\s+map",
            r"\bgenerate\b",
            r"\w+\s*:\s*entity",
            r"\bprocess\b",
        ]
    )
).search


def _getObjectsFromText(lines):
    """
    Returns a dict containing the objects found at the given text buffer
    """
    objects = {}
    lnum = 0
    area = None
    for _line in lines:
        line = re.sub(r"\s*--.*", "", _line)
        for match in _GET_SCOPE(line):
            _dict = match.groupdict()
            if _dict["entity_name"] is not None:
                area = "entity"
            elif _dict["architecture_name"] is not None:
                area = "architecture"
            elif _dict["package_name"] is not None:
                area = "package"
            elif _dict["package_body_name"] is not None:
                area = "package_body"
            else:  # pragma: no cover
                assert False

        matches = []
        if area is None:
            matches += _NO_SCOPE_OBJECTS.finditer(line)
        elif area == "entity":
            matches += _ENTITY_OBJECTS(line)
        elif area == "architecture":
            matches += _ARCH_OBJECTS(line)

        for match in matches:
            for key, value in match.groupdict().items():
                if value is None:
                    continue
                _group_d = match.groupdict()
                index = match.lastindex
                if "port" in _group_d.keys() and _group_d["port"] is not None:
                    index -= 1
                start = match.start(index)
                end = match.end(index)

                # More than 1 declaration can be done in a single line.
                # Must strip white spaces and commas properly
                for submatch in re.finditer(r"(\w+)", value):
                    # Need to decrement the last index because we have a group that
                    # catches the port type (in, out, inout, etc)
                    text = submatch.group(submatch.lastindex)
                    if text not in objects.keys():
                        objects[text] = {}
                    objects[text]["lnum"] = lnum
                    objects[text]["start"] = start + submatch.start(submatch.lastindex)
                    objects[text]["end"] = end + submatch.start(submatch.lastindex)
                    objects[text]["type"] = key
        lnum += 1
        if _SHOULD_END_SCAN(line):
            break

    return objects


def _getUnusedObjects(lines, objects):
    """Generator that yields objects that are only found once at the
    given buffer and thus are considered unused (i.e., we only found
    its declaration"""

    text = ""
    for line in lines:
        text += re.sub(r"\s*--.*", "", line) + " "

    for _object in objects:
        r_len = 0
        single = True
        for _ in re.finditer(r"\b%s\b" % _object, text, flags=re.I):
            r_len += 1
            if r_len > 1:
                single = False
                break
        if single:
            yield _object


__COMMENT_TAG_SCANNER__ = re.compile(
    "|".join([r"\s*--\s*(?P<tag>TODO|FIXME|XXX)\s*:\s*(?P<text>.*)"])
)


def _getCommentTags(lines):
    """
    Generates diags from 'TODO', 'FIXME' and 'XXX' tags
    """
    result = []
    lnum = 0
    for line in lines:
        lnum += 1
        line_lc = line.lower()
        skip_line = True
        for tag in ("todo", "fixme", "xxx"):
            if tag in line_lc:
                skip_line = False
                break
        if skip_line:
            continue

        for match in __COMMENT_TAG_SCANNER__.finditer(line):
            _dict = match.groupdict()
            result += [
                StaticCheckerDiag(
                    line_number=lnum,
                    column_number=match.start(match.lastindex - 1) + 1,
                    severity=DiagType.STYLE_INFO,
                    text="%s: %s" % (_dict["tag"].upper(), _dict["text"]),
                )
            ]
    return result


def _getMiscChecks(objects):
    """
    Get generic code hints (or it should do that sometime in the future...)
    """
    if "library" not in [x["type"] for x in objects.values()]:
        return

    for library, obj in objects.items():
        if obj["type"] != "library":
            continue
        if library == "work":
            yield LibraryShouldBeOmited(
                line_number=obj["lnum"] + 1,
                column_number=obj["start"] + 1,
                library=library,
            )


def getStaticMessages(lines):
    # type: (Tuple[str, ...]) -> Any
    "VHDL static checking"
    objects = _getObjectsFromText(lines)

    result = []

    for _object in _getUnusedObjects(lines, objects.keys()):
        obj_dict = objects[_object]
        result += [
            ObjectIsNeverUsed(
                line_number=obj_dict["lnum"] + 1,
                column_number=obj_dict["start"] + 1,
                object_type=obj_dict["type"],
                object_name=_object,
            )
        ]

    return result + _getCommentTags(lines) + list(_getMiscChecks(objects))


def standalone():  # pragma: no cover
    """
    Standalone entry point
    """
    import sys

    logging.basicConfig(stream=sys.stdout, level=logging.INFO)
    for arg in sys.argv[1:]:
        print(arg)
        lines = [x.decode(errors="ignore") for x in open(arg, mode="rb").readlines()]
        for message in getStaticMessages(lines):
            print(message)
        print("=" * 10)


if __name__ == "__main__":
    standalone()



# This file is part of HDL Checker.
#
# Copyright (c) 2015 - 2019 suoto (Andre Souto)
#
# HDL Checker is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HDL Checker is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with HDL Checker.  If not, see <http://www.gnu.org/licenses/>.
"Handlers for hdl_checker server"

import logging
import os
import os.path as p
import signal
import tempfile
from multiprocessing import Queue
from typing import Any, Dict, List, Optional, Tuple

import bottle  # type: ignore

from hdl_checker import __version__ as version
from hdl_checker import types as t  # pylint: disable=unused-import
from hdl_checker.builders.fallback import Fallback
from hdl_checker.hdl_checker_base import HdlCodeCheckerBase
from hdl_checker.path import Path
from hdl_checker.utils import terminateProcess

_logger = logging.getLogger(__name__)

app = bottle.Bottle()  # pylint: disable=invalid-name


class HdlCodeCheckerServer(HdlCodeCheckerBase):
    """
    HDL Checker project builder class
    """

    def __init__(self, *args, **kwargs):
        # type: (...) -> None
        self._msg_queue = Queue()  # type: Queue[Tuple[str, str]]
        super(HdlCodeCheckerServer, self).__init__(*args, **kwargs)

    def _handleUiInfo(self, message):
        # type: (...) -> Any
        self._msg_queue.put(("info", message))

    def _handleUiWarning(self, message):
        # type: (...) -> Any
        self._msg_queue.put(("warning", message))

    def _handleUiError(self, message):
        # type: (...) -> Any
        self._msg_queue.put(("error", message))

    def getQueuedMessages(self):
        # type: (...) -> Any
        "Returns queued UI messages"
        while not self._msg_queue.empty():  # pragma: no cover
            yield self._msg_queue.get()


def _getServerByProjectFile(project_file):
    # type: (Optional[str]) -> HdlCodeCheckerServer
    """
    Returns the HdlCodeCheckerServer object that corresponds to the given
    project file. If the object doesn't exists yet it gets created and
    then returned
    """
    _logger.debug("project_file: %s", project_file)
    if isinstance(project_file, str) and project_file.lower() == "none":
        project_file = None

    # If there's no project file to use, create a temporary path to use
    if project_file is None:
        root_dir = Path(tempfile.mkdtemp())
    else:
        root_dir = Path(p.dirname(project_file))

    if root_dir not in servers:
        _logger.info("Creating server for %s (root=%s)", project_file, root_dir)
        try:
            project = HdlCodeCheckerServer(root_dir=root_dir)
            if project_file is not None:
                project.setConfig(project_file)
            _logger.debug("Created new project server for '%s'", project_file)
        except (IOError, OSError):
            _logger.info("Failed to create checker, reverting to fallback")
            project = HdlCodeCheckerServer(None)

        servers[root_dir] = project
    return servers[root_dir]


def _exceptionWrapper(func):
    """
    Wraps func to log exception to the standard logger
    """

    def _wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except:  # pragma: no cover
            _logger.exception("Error running '%s'", func.__name__)
            raise

    return _wrapper


def setupSignalHandlers():
    # type: (...) -> Any
    """
    Configures signal handlers that will be called when exiting Python
    shell
    """

    def signalHandler(sig, _):
        "Handle to disable hdl_checker server"
        _logger.info("Handling signal %s", repr(sig))
        import sys

        sys.exit()

    for sig in [signal.SIGTERM, signal.SIGINT]:
        signal.signal(sig, signalHandler)


def _getProjectDiags(project_file):
    # type: (str) -> Any
    """
    Get project specific diagnose
    """
    diags = []  # type: List[str]
    server = _getServerByProjectFile(project_file)

    if isinstance(server.builder, Fallback):
        diags += ["Builder: none"]
    else:
        diags += ["Builder: %s" % server.builder.builder_name]

    return diags


@app.post("/get_diagnose_info")
@_exceptionWrapper
def getDiagnoseInfo():
    # type: (...) -> Any
    """
    Collects misc diagnose info for the clients
    """
    _logger.info("Collecting diagnose info")
    project_file = bottle.request.forms.get("project_file")  # pylint: disable=no-member
    response = ["hdl_checker version: %s" % version, "Server PID: %d" % os.getpid()]

    response += _getProjectDiags(project_file)

    _logger.info("Diagnose info collected:")
    for diag in response:
        _logger.info(" - %s", diag)

    return {"info": response}


@app.post("/get_messages_by_path")
@_exceptionWrapper
def getMessagesByPath():
    # type: (...) -> Any
    """
    Get messages for a given projec_file/path pair
    """
    project_file = bottle.request.forms.get("project_file")  # pylint: disable=no-member
    path = Path(bottle.request.forms.get("path"))  # pylint: disable=no-member
    content = bottle.request.forms.get("content", None)  # pylint: disable=no-member

    _logger.debug(
        "Getting messages for '%s', '%s', %s",
        project_file,
        path,
        "no content" if content is None else "with content",
    )

    server = _getServerByProjectFile(project_file)
    if content is None:
        messages = server.getMessagesByPath(path)
    else:
        messages = server.getMessagesWithText(path, content)

    _logger.info("messages: %s", [x.toDict() for x in messages])

    # Messages at this point need to be serializable so that bottle can send
    # them over
    return {"messages": tuple(x.toDict() for x in messages)}


@app.post("/get_ui_messages")
@_exceptionWrapper
def getUiMessages():
    # type: (...) -> Any
    """
    Get messages for a given projec_file/path pair
    """

    project_file = bottle.request.forms.get("project_file")  # pylint: disable=no-member
    server = _getServerByProjectFile(project_file)

    ui_messages = list(server.getQueuedMessages())

    if not ui_messages:  # pragma: no cover
        _logger.debug("Project '%s' has no UI messages", project_file)
    else:  # pragma: no cover
        _logger.info("Project '%s' UI messages:", project_file)

    for msg in ui_messages:  # pragma: no cover
        _logger.info(msg)

    response = {"ui_messages": ui_messages}

    return response


@app.post("/rebuild_project")
@_exceptionWrapper
def rebuildProject():
    # type: (...) -> Any
    """
    Rebuilds the current project
    """
    _logger.info("Rebuilding project")
    project_file = bottle.request.forms.get("project_file")  # pylint: disable=no-member
    server = _getServerByProjectFile(project_file)
    server.clean()
    _logger.debug("Removing and recreating server object")
    del servers[project_file]
    _getServerByProjectFile(project_file)


@app.post("/shutdown")
@_exceptionWrapper
def shutdownServer():
    # type: (...) -> Any
    """
    Terminates the current process to shutdown the server
    """
    _logger.info("Shutting down server")
    terminateProcess(os.getpid())


@app.post("/get_dependencies")
@_exceptionWrapper
def getDependencies():
    # type: (...) -> Any
    """
    Returns the direct dependencies of a given source path
    """
    project_file = bottle.request.forms.get("project_file")  # pylint: disable=no-member
    path = Path(bottle.request.forms.get("path"))  # pylint: disable=no-member

    _logger.debug("Getting dependencies for '%s', '%s'", project_file, path)

    server = _getServerByProjectFile(project_file)
    content = []
    for dependency in server.database.getDependenciesByPath(path):
        content.append(
            "%s.%s" % (dependency.library.display_name, dependency.name.display_name)
        )

    _logger.debug("Found %d dependencies", len(content))

    return {"dependencies": content}


@app.post("/get_build_sequence")
@_exceptionWrapper
def getBuildSequence():
    # type: (...) -> Any
    """
    Returns the build sequence of a given source path
    """
    project_file = bottle.request.forms.get("project_file")  # pylint: disable=no-member
    path = Path(bottle.request.forms.get("path"))  # pylint: disable=no-member

    _logger.debug("Getting build sequence for '%s', '%s'", project_file, path)

    server = _getServerByProjectFile(project_file)
    server._updateConfigIfNeeded()

    return {
        "sequence": tuple(
            "%s (library: %s)" % (path, library)
            for library, path in server.database.getBuildSequence(
                path, server.builder.builtin_libraries
            )
        )
    }


#  We'll store a dict to store differents hdl_checker objects
servers = {}  # type: Dict[Path, HdlCodeCheckerServer] # pylint: disable=invalid-name
setupSignalHandlers()



# This file is part of HDL Checker.
#
# Copyright (c) 2015 - 2019 suoto (Andre Souto)
#
# HDL Checker is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HDL Checker is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with HDL Checker.  If not, see <http://www.gnu.org/licenses/>.

# pylint: disable=function-redefined
# pylint: disable=invalid-name
# pylint: disable=missing-docstring
# pylint: disable=protected-access
# pylint: disable=useless-object-inheritance

import json
import logging
import os.path as p
from pprint import pformat

from hdl_checker.tests import TestCase, getTestTempPath

from hdl_checker.parser_utils import flattenConfig, getIncludedConfigs
from hdl_checker.path import Path
from hdl_checker.types import BuildFlagScope, FileType
from hdl_checker.utils import removeIfExists

_logger = logging.getLogger(__name__)

TEST_TEMP_PATH = getTestTempPath(__name__)
TEST_PROJECT = p.join(TEST_TEMP_PATH, "test_project")

TEST_CONFIG_PARSER_SUPPORT_PATH = p.join(TEST_TEMP_PATH, "test_config_parser")


def _path(*args):
    # type: (str) -> str
    "Helper to reduce foorprint of p.join(TEST_TEMP_PATH, *args)"
    return p.join(TEST_TEMP_PATH, *args)


def _Path(*args):
    # type: (str) -> Path
    "Helper to reduce foorprint of p.join(TEST_TEMP_PATH, *args)"
    return Path(_path(*args))


def json_dump(obj, stream):
    _logger.info("JSON dump: %s:\n%s", stream, pformat(obj))
    json.dump(obj, stream)


class _ConfigDict(object):
    def __init__(self):
        self.sources = []
        self.flags = {}
        self.include = []
        for lang in FileType:
            self.flags[lang] = {}
            for scope in BuildFlagScope:
                self.flags[lang][scope] = []

    def toDict(self):
        d = {"sources": tuple(self.sources), "include": tuple(self.include)}

        for lang in FileType:
            d[lang.value] = {"flags": {}}
            for scope in BuildFlagScope:
                d[lang.value]["flags"][scope.value] = tuple(self.flags[lang][scope])
        return d


class TestConfigHandlers(TestCase):
    maxDiff = None

    def test_direct_inclusion(self):
        incl_0 = _path("incl_0.json")
        incl_1 = _path("incl_1.json")
        incl_2 = _path("incl_2.json")
        incl_3 = _path("incl_3.json")
        #  incl_4 = _path('incl_4.json')

        search_paths = (incl_0, "incl_1.json")

        # Direct inclusion
        json_dump({"include": (incl_1,), "name": "incl_0"}, open(incl_0, "w"))
        # Direct multiple inclusion
        json_dump({"include": (incl_2, incl_3), "name": "incl_1"}, open(incl_1, "w"))
        # No inclusion (enpoints)
        json_dump({"name": "incl_2"}, open(incl_2, "w"))
        json_dump({"name": "incl_3"}, open(incl_3, "w"))

        result = list(getIncludedConfigs(search_paths, TEST_TEMP_PATH))

        _logger.info("Result:\n%s", pformat(result))
        self.assertCountEqual(
            result,
            (
                (TEST_TEMP_PATH, {"name": "incl_0"}),
                (TEST_TEMP_PATH, {"name": "incl_1"}),
                (TEST_TEMP_PATH, {"name": "incl_2"}),
                (TEST_TEMP_PATH, {"name": "incl_3"}),
            ),
        )

    def test_recursive_inclusion(self):
        incl_0 = _path("incl_0.json")
        incl_1 = _path("incl_1.json")

        search_paths = (incl_0,)

        json_dump({"include": (incl_1,), "name": "incl_0"}, open(incl_0, "w"))
        json_dump({"include": (incl_0,), "name": "incl_1"}, open(incl_1, "w"))

        result = list(getIncludedConfigs(search_paths, TEST_TEMP_PATH))

        _logger.info("Result:\n%s", pformat(result))
        self.assertCountEqual(
            result,
            (
                (TEST_TEMP_PATH, {"name": "incl_0"}),
                (TEST_TEMP_PATH, {"name": "incl_1"}),
            ),
        )

    def test_ignores_non_existing_files(self):
        incl_0 = _path("incl_0.json")
        incl_1 = _path("incl_1.json")

        search_paths = (incl_0, incl_1)

        removeIfExists(incl_0)
        json_dump({"name": "incl_1"}, open(incl_1, "w"))

        result = list(getIncludedConfigs(search_paths, TEST_TEMP_PATH))

        _logger.info("Result:\n%s", pformat(result))
        self.assertCountEqual(result, ((TEST_TEMP_PATH, {"name": "incl_1"}),))

    def test_ignores_json_decoding_errors(self):
        search_paths = (_path("incl_0.json"),)

        open(_path("incl_0.json"), "w").write("hello")

        result = list(getIncludedConfigs(search_paths, TEST_TEMP_PATH))

        _logger.info("Result:\n%s", pformat(result))
        self.assertCountEqual(result, ())

    def test_includes_relative_paths(self):
        incl_0 = _path("incl_0.json")
        incl_1 = _path("incl_1.json")
        incl_2 = _path("incl_2.json")

        search_paths = (incl_0, "incl_1.json")

        # Direct inclusion
        json_dump({"include": ("incl_1.json",), "name": "incl_0"}, open(incl_0, "w"))
        # Direct multiple inclusion
        json_dump({"include": ("incl_2.json",), "name": "incl_1"}, open(incl_1, "w"))
        # No inclusion (enpoints)
        json_dump({"name": "incl_2"}, open(incl_2, "w"))

        result = list(getIncludedConfigs(search_paths, TEST_TEMP_PATH))

        _logger.info("Result:\n%s", pformat(result))
        self.assertCountEqual(
            result,
            (
                (TEST_TEMP_PATH, {"name": "incl_0"}),
                (TEST_TEMP_PATH, {"name": "incl_1"}),
                (TEST_TEMP_PATH, {"name": "incl_2"}),
            ),
        )

    def test_flatten_config_and_preserve_scopes(self):
        incl_0 = _path("incl_0.json")
        incl_1 = _path("incl_1.json")
        #  incl_2 = _path("incl_2.json")

        incl_0_cfg = _ConfigDict()

        incl_0_cfg.include += [incl_1]

        incl_0_cfg.sources += [
            "src_0_0.vhd",
            "src_0_1.v",
            "src_0_2.sv",
            #  ("src_0_3.vhd", {"library": "some_library"}),
            #  ("src_0_4.vhd", {"flags": ("some_flag",)}),
            #  ("src_0_5.vhd", {"library": "lib", "flags": ("and_flag",)}),
            ("src_0_3.vhd", {"library": "l_0_3"}),
            ("src_0_4.vhd", {"flags": ("f_0_4",)}),
            ("src_0_5.vhd", {"library": "l_0_5", "flags": ("f_0_5",)}),
        ]

        incl_0_cfg.flags = {
            FileType.vhdl: {
                BuildFlagScope.all: ["vhdl/0/glob"],
                BuildFlagScope.dependencies: ["vhdl/0/deps"],
                BuildFlagScope.single: ["vhdl/0/single"],
            },
            FileType.verilog: {
                BuildFlagScope.all: ["verilog/0/glob"],
                BuildFlagScope.dependencies: ["verilog/0/deps"],
                BuildFlagScope.single: ["verilog/0/single"],
            },
            FileType.systemverilog: {
                BuildFlagScope.all: ["systemverilog/0/glob"],
                BuildFlagScope.dependencies: ["systemverilog/0/deps"],
                BuildFlagScope.single: ["systemverilog/0/single"],
            },
        }

        incl_1_cfg = _ConfigDict()

        incl_1_cfg.sources += [
            _path("src_1_0.vhd"),
            "src_1_1.v",
            "src_1_2.sv",
            ("src_1_3.vhd", {"library": "l_1_3"}),
            (_path("src_1_4.vhd"), {"flags": ("f_1_4",)}),
            ("src_1_5.vhd", {"library": "l_1_5", "flags": ("f_1_5",)}),
        ]

        incl_1_cfg.flags = {
            FileType.vhdl: {
                BuildFlagScope.all: ["vhdl/1/glob"],
                BuildFlagScope.dependencies: ["vhdl/1/deps"],
                BuildFlagScope.single: ["vhdl/1/single"],
            },
            FileType.verilog: {
                BuildFlagScope.all: ["verilog/1/glob"],
                BuildFlagScope.dependencies: ["verilog/1/deps"],
                BuildFlagScope.single: ["verilog/1/single"],
            },
            FileType.systemverilog: {
                BuildFlagScope.all: ["systemverilog/1/glob"],
                BuildFlagScope.dependencies: ["systemverilog/1/deps"],
                BuildFlagScope.single: ["systemverilog/1/single"],
            },
        }

        json_dump(incl_0_cfg.toDict(), open(incl_0, "w"))
        json_dump(incl_1_cfg.toDict(), open(incl_1, "w"))

        result = list(flattenConfig(incl_0_cfg.toDict(), TEST_TEMP_PATH))

        _logger.info("Result:\n%s", pformat(result))

        self.assertCountEqual(
            result,
            (
                (
                    _Path("src_0_0.vhd"),
                    None,
                    ("vhdl/0/glob", "vhdl/0/single"),
                    ("vhdl/0/glob", "vhdl/0/deps"),
                ),
                (
                    _Path("src_0_1.v"),
                    None,
                    ("verilog/0/glob", "verilog/0/single"),
                    ("verilog/0/glob", "verilog/0/deps"),
                ),
                (
                    _Path("src_0_2.sv"),
                    None,
                    ("systemverilog/0/glob", "systemverilog/0/single"),
                    ("systemverilog/0/glob", "systemverilog/0/deps"),
                ),
                (
                    _Path("src_0_3.vhd"),
                    "l_0_3",
                    ("vhdl/0/glob", "vhdl/0/single"),
                    ("vhdl/0/glob", "vhdl/0/deps"),
                ),
                (
                    _Path("src_0_4.vhd"),
                    None,
                    ("vhdl/0/glob", "vhdl/0/single", "f_0_4"),
                    ("vhdl/0/glob", "vhdl/0/deps"),
                ),
                (
                    _Path("src_0_5.vhd"),
                    "l_0_5",
                    ("vhdl/0/glob", "vhdl/0/single", "f_0_5"),
                    ("vhdl/0/glob", "vhdl/0/deps"),
                ),
                (
                    _Path("src_1_0.vhd"),
                    None,
                    ("vhdl/1/glob", "vhdl/1/single"),
                    ("vhdl/1/glob", "vhdl/1/deps"),
                ),
                (
                    _Path("src_1_1.v"),
                    None,
                    ("verilog/1/glob", "verilog/1/single"),
                    ("verilog/1/glob", "verilog/1/deps"),
                ),
                (
                    _Path("src_1_2.sv"),
                    None,
                    ("systemverilog/1/glob", "systemverilog/1/single"),
                    ("systemverilog/1/glob", "systemverilog/1/deps"),
                ),
                (
                    _Path("src_1_3.vhd"),
                    "l_1_3",
                    ("vhdl/1/glob", "vhdl/1/single"),
                    ("vhdl/1/glob", "vhdl/1/deps"),
                ),
                (
                    _Path("src_1_4.vhd"),
                    None,
                    ("vhdl/1/glob", "vhdl/1/single", "f_1_4"),
                    ("vhdl/1/glob", "vhdl/1/deps"),
                ),
                (
                    _Path("src_1_5.vhd"),
                    "l_1_5",
                    ("vhdl/1/glob", "vhdl/1/single", "f_1_5"),
                    ("vhdl/1/glob", "vhdl/1/deps"),
                ),
            ),
        )



# This file is part of HDL Checker.
#
# Copyright (c) 2015 - 2019 suoto (Andre Souto)
#
# HDL Checker is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HDL Checker is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with HDL Checker.  If not, see <http://www.gnu.org/licenses/>.
'Exceptions raised by hdl_checker'

class HdlCheckerBaseException(Exception):
    """
    Base class for exceptions raise by hdl_checker
    """

class SanityCheckError(HdlCheckerBaseException):
    """
    Exception raised when a builder fails to execute its sanity check
    """

    def __init__(self, builder, msg):
        self._msg = msg
        self.builder = builder
        super(SanityCheckError, self).__init__()

    def __str__(self): # pragma: no cover
        return "Failed to create builder '%s' with message '%s'" % \
                (self.builder, self._msg)

class UnknownParameterError(HdlCheckerBaseException):
    """
    Exception raised when an unknown parameter is found in a
    configuration file
    """

    def __init__(self, parameter):
        self._parameter = parameter
        super(UnknownParameterError, self).__init__()

    def __str__(self): # pragma: no cover
        return "Unknown parameter '%s'" % self._parameter




# This file is part of HDL Checker.
#
# Copyright (c) 2015 - 2019 suoto (Andre Souto)
#
# HDL Checker is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HDL Checker is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with HDL Checker.  If not, see <http://www.gnu.org/licenses/>.

# pylint: disable=missing-docstring
# pylint: disable=useless-object-inheritance

import logging
import os
import os.path as p
import shutil
import subprocess as subp
import time
from multiprocessing import Queue
from typing import Any, Callable, Dict, Iterable, List, Optional, Tuple, Union

import mock
import six
from parameterized import parameterized_class  # type: ignore

from hdl_checker import exceptions
from hdl_checker.builders.base_builder import BaseBuilder
from hdl_checker.hdl_checker_base import HdlCodeCheckerBase
from hdl_checker.parsers.elements.dependency_spec import DependencySpec
from hdl_checker.parsers.elements.identifier import Identifier
from hdl_checker.path import Path
from hdl_checker.types import FileType
from hdl_checker.utils import ON_WINDOWS, removeDuplicates, samefile

_logger = logging.getLogger(__name__)


MockDep = Union[Tuple[str], Tuple[str, str]]


class StandaloneProjectBuilder(HdlCodeCheckerBase):
    "Class for testing HdlCodeCheckerBase"
    _msg_queue = Queue()  # type: Queue[Tuple[str, str]]
    _ui_handler = logging.getLogger("UI")

    def _handleUiInfo(self, message):
        self._msg_queue.put(("info", message))
        self._ui_handler.info("[UI INFO]: %s", message)

    def _handleUiWarning(self, message):
        self._msg_queue.put(("warning", message))
        self._ui_handler.info("[UI WARNING]: %s", message)

    def _handleUiError(self, message):
        self._msg_queue.put(("error", message))
        self._ui_handler.info("[UI ERROR]: %s", message)

    def getUiMessages(self):
        while not self._msg_queue.empty():
            yield self._msg_queue.get()


class SourceMock(object):
    _logger = logging.getLogger("SourceMock")
    base_path = ""

    def __init__(
        self,
        design_units,  # type: Iterable[Dict[str, str]]
        library=None,  # type: str
        dependencies=None,  # type: Iterable[MockDep]
        filename=None,  # type: Optional[str]
    ):

        self._design_units = list(design_units or [])

        if filename is not None:
            self._filename = Path(p.join(self.base_path, filename))
        else:
            library = "lib_not_set" if library is None else library
            self._filename = Path(
                p.join(
                    self.base_path,
                    library,
                    "_{}.vhd".format(self._design_units[0]["name"]),
                )
            )

        self.filetype = FileType.fromPath(self.filename)
        #  self.abspath = p.abspath(self.filename)
        self.flags = []  # type: ignore

        self.library = library
        self._dependencies = []  # type: List[DependencySpec]
        for dep_spec in dependencies or []:
            _name = dep_spec[0]
            _library = "work"

            try:
                _library, _name = dep_spec  # type: ignore
            except ValueError:
                pass

            self._dependencies.append(
                DependencySpec(
                    self._filename,
                    Identifier(_name, False),
                    Identifier(_library, False),
                )
            )

        self._createMockFile()

    @property
    def filename(self):
        # type: () -> Path
        return self._filename

    def getLibraries(self):
        return removeDuplicates([x.library for x in self._dependencies])

    def _createMockFile(self):
        # type: () -> None
        self._logger.debug("Creating mock file: %s", self)
        libs = self.getLibraries()

        lines = []

        for lib in libs:
            lines.append("library {0};".format(lib.display_name))

        for dependency in self._dependencies:
            lines.append(
                "use {0}.{1};".format(
                    dependency.library.display_name, dependency.name.display_name
                )
            )

        # Separate if there was libraries already added
        if lines:
            lines.append("")

        for design_unit in self._design_units:
            type_ = design_unit["type"]
            name = design_unit["name"]

            lines.append("{0} {1} is".format(type_, name))
            lines.append("")
            lines.append("end {0} {1};".format(type_, name))
            lines.append("")

        for i, line in enumerate(lines, 1):
            self._logger.debug("%2d | %s", i, line)

        try:
            os.makedirs(p.dirname(self.filename.name))
        except OSError:
            pass

        with open(self.filename.name, "w") as fd:
            fd.write("\n".join(lines))

    def __repr__(self):
        return (
            "{}(library='{}', design_units={}, dependencies={}, "
            "filename={}".format(
                self.__class__.__name__,
                self.library,
                self._design_units,
                self._dependencies,
                self.filename,
            )
        )

    def getmtime(self):
        return p.getmtime(self.filename)

    def getDesignUnits(self):
        return self._design_units

    def getDependencies(self):
        return self._dependencies

    def getRawSourceContent(self):
        return open(self.filename).read()


class MockBuilder(BaseBuilder):  # pylint: disable=abstract-method
    _logger = logging.getLogger("MockBuilder")
    builder_name = "mock_builder"
    file_types = (FileType.vhdl,)

    def __init__(self, work_folder, *args, **kwargs):
        # type: (...) -> None
        self._work_folder = work_folder  # type: Path
        if not p.exists(self._work_folder.name):
            os.makedirs(self._work_folder.name)

        super(MockBuilder, self).__init__(work_folder, *args, **kwargs)

    def _makeRecords(self, _):  # pragma: no cover
        return []

    def _shouldIgnoreLine(self, line):  # pragma: no cover
        return True

    def _checkEnvironment(self):
        return

    @staticmethod
    def isAvailable():
        return True

    def _buildSource(self, path, library, flags=None):
        self._logger.debug(
            "Building path=%s, library=%s, flags=%s", path, library, flags
        )
        return [], []

    def _createLibrary(self, library):  # pylint: disable=unused-argument
        pass

    def _parseBuiltinLibraries(self):
        # type: (...) -> Any
        return (
            Identifier("ieee", case_sensitive=False),
            Identifier("std", case_sensitive=False),
        )


class FailingBuilder(MockBuilder):  # pylint: disable=abstract-method
    _logger = logging.getLogger("FailingBuilder")
    builder_name = "FailingBuilder"

    def _checkEnvironment(self):
        raise exceptions.SanityCheckError(self.builder_name, "Fake error")


disableVunit = mock.patch("hdl_checker.builder_utils.foundVunit", lambda: False)


class PatchBuilder(object):
    def __init__(self, meth=None):
        def getBuilderByName(name):
            "Returns the builder class given a string name"
            from hdl_checker.builders.msim import MSim
            from hdl_checker.builders.ghdl import GHDL
            from hdl_checker.builders.xvhdl import XVHDL
            from hdl_checker.builders.fallback import Fallback

            # Check if the builder selected is implemented and create the
            # builder attribute
            _logger.info("Getting builder class for %s", name)
            if name == MockBuilder.builder_name:
                return MockBuilder
            if name == "msim":
                return MSim
            if name == "xvhdl":
                return XVHDL
            if name == "ghdl":
                return GHDL

            return Fallback

        self.meth = meth
        self.patches = (
            mock.patch(
                "hdl_checker.hdl_checker_base.getWorkingBuilders",
                side_effect=[iter((MockBuilder,))],
            ),
            mock.patch("hdl_checker.hdl_checker_base.getBuilderByName", getBuilderByName),
            disableVunit,
        )

    def __enter__(self):
        _logger.info("Starting patches")
        list(x.start() for x in self.patches)

    def __exit__(self, *args, **kwargs):
        _logger.info("Stopping patches")
        list(x.stop() for x in self.patches)

    def __call__(self, *args, **kwargs):
        self.__enter__()
        try:
            if self.meth is not None:
                yield self.meth()
            else:
                yield None
        except:
            _logger.exception("Failed to run %s", self.meth)
            raise
        finally:
            self.__exit__()


def sanitizePath(*args):
    """
    Join args with pathsep, gets the absolute path and normalizes
    """
    return p.normpath(
        p.abspath(p.join(*args))  # pylint: disable=no-value-for-parameter
    )


def assertSameFile(it):  # pylint: disable=invalid-name
    def wrapper(first, second, msg=None):
        try:
            os.stat(first)
        except TypeError:
            it.fail("Invalid first argument of type {}".format(type(first)))
        try:
            os.stat(second)
        except TypeError:
            it.fail("Invalid first argument of type {}".format(type(second)))
        if not samefile(p.abspath(first), p.abspath(second)):
            _msg = "" if msg is None else "{}\n".format(msg)
            it.fail(
                "{}Paths '{}' and '{}' differ".format(
                    _msg, p.abspath(first), p.abspath(second)
                )
            )

    return wrapper


def assertCountEqual(it):  # pylint: disable=invalid-name

    assert six.PY2, "Only needed on Python2"

    def wrapper(first, second, msg=None):
        temp = list(second)  # make a mutable copy
        not_found = []
        for elem in first:
            try:
                temp.remove(elem)
            except ValueError:
                not_found.append(elem)

        error_details = []

        if not_found:
            error_details += [
                "Second list is missing item {}".format(x) for x in not_found
            ]

        error_details += ["First list is missing item {}".format(x) for x in temp]

        if error_details:
            # Add user message at the top
            error_details = [msg] + error_details
            error_details += ["", "Lists {} and {} differ".format(first, second)]
            it.fail("\n".join([str(x) for x in error_details]))

    return wrapper


def writeListToFile(filename, _list):  # pragma: no cover
    "Well... writes '_list' to 'filename'. This is for testing only"
    # Wait a little bit to force the timestamp rad via os.path.getmtime to
    # change
    time.sleep(0.1)
    open(filename, mode="w").write("\n".join([str(x) for x in _list]))

    for i, line in enumerate(_list):
        _logger.debug("%2d | %s", i + 1, line)


if not ON_WINDOWS:
    TEST_ENVS = {
        "ghdl": os.environ["GHDL_PATH"],
        "msim": os.environ["MODELSIM_PATH"],
        "xvhdl": os.environ["XSIM_PATH"],
        "fallback": None,
    }
else:
    TEST_ENVS = {"fallback": None}


def parametrizeClassWithBuilders(cls):
    cls.assertSameFile = assertSameFile(cls)

    keys = ["builder_name", "builder_path"]
    values = []
    for name, path in TEST_ENVS.items():
        values += [(name, path)]

    return parameterized_class(keys, values)(cls)


def getTestTempPath(name):
    # type: (str) -> str
    name = name.replace(".", "_")
    path = p.abspath(p.join(os.environ["TOX_ENV_DIR"], "tmp", name))

    # Create a path for each test to allow running multiple tests concurrently
    i = 0
    while p.exists(path):
        path = p.abspath(p.join(os.environ["TOX_ENV_DIR"], "tmp", "%s_%d" % (name, i)))
        i += 1

    if not p.exists(path):
        os.makedirs(path)
    return path


def setupTestSuport(path):
    # type: (str) -> None
    """Copy contents of .ci/test_support_path/ to the given path"""
    _logger.info("Setting up test support at %s", path)

    test_support_path = os.environ["CI_TEST_SUPPORT_PATH"]

    paths_to_copy = os.listdir(test_support_path)

    # Create the parent directory
    if not p.exists(p.dirname(path)):
        _logger.info("Creating %s", p.dirname(path))
        os.makedirs(p.dirname(path))

    for path_to_copy in paths_to_copy:
        src = p.join(test_support_path, path_to_copy)
        dest = p.join(path, path_to_copy)

        if p.exists(dest):
            _logger.info("Destination path %s already exists, removing it", dest)
            shutil.rmtree(dest)

        _logger.info("Copying %s to %s", src, dest)
        shutil.copytree(src, dest)


def logIterable(msg, iterable, func):
    # type: (str, Iterable[Any], Callable) -> None
    func(msg)
    for i, item in enumerate(iterable, 1):
        func("- {:2d} {}".format(i, item))


if six.PY2:
    from unittest2 import TestCase as _TestCase

    class TestCase(_TestCase):
        def __new__(cls, *args, **kwargs):
            result = super(_TestCase, cls).__new__(cls, *args, **kwargs)
            result.assertCountEqual = assertCountEqual(result)
            return result


else:
    from unittest2 import TestCase  # type: ignore



# This file is part of HDL Checker.
#
# Copyright (c) 2015 - 2019 suoto (Andre Souto)
#
# HDL Checker is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HDL Checker is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with HDL Checker.  If not, see <http://www.gnu.org/licenses/>.

# pylint: disable=function-redefined
# pylint: disable=invalid-name
# pylint: disable=missing-docstring
# pylint: disable=protected-access
# pylint: disable=useless-object-inheritance

import json
import logging
import os.path as p
import time
from pprint import pformat
from typing import Any, Dict, Iterable, Set, Tuple

from mock import patch

from hdl_checker.tests import (
    SourceMock,
    TestCase,
    getTestTempPath,
    logIterable,
    setupTestSuport,
)

from hdl_checker.database import Database
from hdl_checker.diagnostics import DependencyNotUnique, PathNotInProjectFile
from hdl_checker.parsers.elements.dependency_spec import DependencySpec
from hdl_checker.parsers.elements.design_unit import DesignUnitType, VhdlDesignUnit
from hdl_checker.parsers.elements.identifier import Identifier
from hdl_checker.path import Path, TemporaryPath
from hdl_checker.serialization import StateEncoder, jsonObjectHook
from hdl_checker.types import BuildFlagScope, FileType

_logger = logging.getLogger(__name__)

TEST_TEMP_PATH = getTestTempPath(__name__)
TEST_PROJECT = p.join(TEST_TEMP_PATH, "test_project")

TEST_CONFIG_PARSER_SUPPORT_PATH = p.join(TEST_TEMP_PATH, "test_config_parser")


class _SourceMock(SourceMock):
    base_path = TEST_TEMP_PATH


class _Database(Database):
    def configure(self, root_config, root_path):
        # type: (Dict[str, Any], str) -> None
        _logger.info("Updating config from\n%s", pformat(root_config))
        super(_Database, self).configure(root_config, root_path)

        _logger.debug("State after updating:")

        _logger.debug("- %d design units:", len(self.design_units))
        for unit in self.design_units:
            _logger.debug("  - %s", unit)

        _logger.debug("- %d paths:", len(self._paths))
        for path in self._paths:
            timestamp = self._parse_timestamp[path]
            dependencies = self._dependencies_map.get(
                path, set()
            )  # type: Set[DependencySpec]
            _logger.debug("  - Path: %s (%f)", path, timestamp)
            _logger.debug("    - library:      : %s", self._library_map.get(path, "?"))
            _logger.debug("    - flags:        : %s", self._flags_map.get(path, "-"))
            _logger.debug("    - %d dependencies:", len(dependencies))
            for dependency in dependencies:
                _logger.debug("      - %s", dependency)

    def __jsonEncode__(self):
        state = super(_Database, self).__jsonEncode__()
        state["__class__"] = "Database"  # super(_Database, self).__class__.__name__
        return state

    def _configFromSources(self, sources, root_path):
        "Creates the dict describing the sources and calls self.configure on it"
        config = []
        for source in sources:
            info = {}
            if source.library:
                info["library"] = source.library
            if source.flags:
                info["flags"] = source.flags

            # Yield a string only or a (str, dict) tuple to test both conversions
            # (this mimics the supported JSON scheme)
            if info:
                config.append((source.filename.name, info))
            else:
                config.append(source.filename.name)

        self.configure({"sources": config}, root_path)

    def test_getDependenciesUnits(self, path):
        # type: (Path) -> Iterable[Tuple[Identifier, Identifier]]
        _msg = []
        for library, name in super(_Database, self).getDependenciesUnits(path):
            yield getattr(library, "name", None), name.name
            _msg.append((library, name))

        _logger.debug("getDependenciesUnits('%s') => %s", path, _msg)

    def test_getBuildSequence(self, path):
        # type: (Path) -> Iterable[Tuple[Identifier, Path]]
        _msg = []
        for library, build_path in super(_Database, self).getBuildSequence(path):
            yield library, build_path
            _msg.append((getattr(library, "name", None), build_path.name))

        _logger.debug("getBuildSequence('%s') => %s", path, _msg)

    def test_reportCacheInfo(self):
        # Print cache stats
        lines = []
        for attr_name in dir(self):
            if attr_name.startswith("__"):
                continue
            try:
                meth = getattr(getattr(self, attr_name), "cache_info")
                lines.append(" - Method '{}': {}".format(attr_name, meth()))
            except AttributeError:
                pass
        if lines:
            _logger.info("Cache info for %s", self)
            for line in lines:
                _logger.info(line)
        else:
            _logger.info("No cache info for %s", self)


def _path(*args):
    # type: (str) -> str
    "Helper to reduce foorprint of p.join(TEST_TEMP_PATH, *args)"
    return p.join(TEST_TEMP_PATH, *args)


def _Path(*args):
    # type: (str) -> Path
    return Path(_path(*args))


class TestDatabase(TestCase):
    maxDiff = None

    def setUp(self):
        # type: (...) -> Any
        _logger.info("Setting up %s", self)
        setupTestSuport(TEST_TEMP_PATH)
        self.database = _Database()

    def test_accepts_empty_sources_list(self):
        # type: (...) -> Any
        self.database.configure(dict(), TEST_TEMP_PATH)

        #  self.assertEqual(self.database.builder_name, None)
        self.assertCountEqual(self.database.paths, ())
        any_path = _Path("any.vhd")
        # Make sure the path exists
        open(any_path.name, "w").close()
        self.assertEqual(
            self.database.getLibrary(any_path), Identifier("not_in_project", False)
        )
        self.assertEqual(self.database.getFlags(any_path), ())

    def test_accepts_empty_dict(self):
        # type: (...) -> Any
        self.database.configure({}, TEST_TEMP_PATH)

        #  self.assertEqual(self.database.builder_name, None)
        self.assertCountEqual(self.database.paths, ())
        #  self.assertEqual(self.database.getLibrary(Path("any")), None)
        self.assertEqual(self.database.getFlags(Path("any")), ())

    def test_accepts_basic_structure(self):
        # type: (...) -> Any
        _SourceMock(
            filename=_path("foo.vhd"),
            design_units=[{"name": "entity_a", "type": "entity"}],
            dependencies={("some_entity",)},
        )

        _SourceMock(
            filename=_path("oof.vhd"),
            design_units=[{"name": "entity_b", "type": "entity"}],
            dependencies={("some_entity",)},
        )

        self.database.configure(
            {
                "sources": [
                    (_path("foo.vhd"), {"library": "bar", "flags": ("baz", "flag")}),
                    (
                        _path("oof.vhd"),
                        {"library": "ooflib", "flags": ("oofflag0", "oofflag1")},
                    ),
                ],
                FileType.vhdl.value: {
                    "flags": {
                        BuildFlagScope.single.value: ("vhdl", "single"),
                        BuildFlagScope.dependencies.value: ("vhdl", "dependencies"),
                        BuildFlagScope.all.value: ("vhdl", "all"),
                    }
                },
                FileType.verilog.value: {
                    "flags": {
                        BuildFlagScope.single.value: ("verilog", "single"),
                        BuildFlagScope.dependencies.value: ("verilog", "dependencies"),
                        BuildFlagScope.all.value: ("verilog", "all"),
                    }
                },
                FileType.systemverilog.value: {
                    "flags": {
                        BuildFlagScope.single.value: ("systemverilog", "single"),
                        BuildFlagScope.dependencies.value: (
                            "systemverilog",
                            "dependencies",
                        ),
                        BuildFlagScope.all.value: ("systemverilog", "all"),
                    }
                },
            },
            TEST_TEMP_PATH,
        )

        foo_path = _Path("foo.vhd")
        oof_path = _Path("oof.vhd")

        self.assertCountEqual(self.database.paths, (foo_path, oof_path))
        self.assertEqual(self.database.getLibrary(foo_path), Identifier("bar", False))
        self.assertEqual(
            self.database.getLibrary(oof_path), Identifier("ooflib", False)
        )

        self.assertEqual(
            self.database.getFlags(foo_path, BuildFlagScope.single),
            ("vhdl", "all", "vhdl", "single", "baz", "flag"),
        )
        self.assertEqual(
            self.database.getFlags(foo_path),
            ("vhdl", "all", "vhdl", "single", "baz", "flag"),
        )

        #  self.assertEqual(self.database.getFlags(oof_path), ("oofflag0", "oofflag1"))

        logIterable(
            "Design units:", self.database.getDesignUnitsByPath(foo_path), _logger.info
        )

        self.assertCountEqual(
            self.database.getDesignUnitsByPath(foo_path),
            {
                VhdlDesignUnit(
                    owner=foo_path,
                    name="entity_a",
                    type_=DesignUnitType.entity,
                    locations={(3, None)},
                )
            },
        )

        # VHDL world, should find regardless of lower or upper case
        self.assertCountEqual(
            self.database.getPathsByDesignUnit(
                VhdlDesignUnit(Path(""), DesignUnitType.entity, "entity_a")
            ),
            {foo_path},
        )

        self.assertCountEqual(
            self.database.getPathsByDesignUnit(
                VhdlDesignUnit(Path(""), DesignUnitType.entity, "ENTITY_A")
            ),
            {foo_path},
        )

    def test_update_info_if_source_changed(self):
        # type: (...) -> Any
        self.test_accepts_basic_structure()

        # Make sure the env is sane before actually testing
        self.assertTrue(p.exists(_path("foo.vhd")))
        timestamp = p.getmtime(_path("foo.vhd"))

        time.sleep(0.5)

        # SourceMock object writes a dummy VHDL, that should cause the
        # timestamp to change
        _SourceMock(
            filename=_path("foo.vhd"),
            design_units=[{"name": "entity_b", "type": "entity"}],
            dependencies=[],
        )

        self.assertNotEqual(
            timestamp, p.getmtime(_path("foo.vhd")), "Timestamp should've changed"
        )

        self.assertCountEqual(
            self.database.getDesignUnitsByPath(_Path("foo.vhd")),
            [
                VhdlDesignUnit(
                    owner=_Path("foo.vhd"),
                    name="entity_b",
                    type_=DesignUnitType.entity,
                    locations={(0, None)},
                )
            ],
        )

        self.assertCountEqual(
            self.database.getPathsByDesignUnit(
                VhdlDesignUnit(Path(""), DesignUnitType.entity, "entity_a")
            ),
            (),
        )

        oof_path = _Path("oof.vhd")

        self.assertCountEqual(
            self.database.getPathsByDesignUnit(
                VhdlDesignUnit(Path(""), DesignUnitType.entity, "entity_B")
            ),
            [_Path("foo.vhd"), oof_path],
        )

    def test_update_path_library(self):
        # type: (...) -> Any
        sources = {
            _SourceMock(
                filename=_path("file_0.vhd"),
                library="some_library",
                design_units=[{"name": "some_package", "type": "package"}],
                dependencies=(
                    ("work", "relative_dependency"),
                    ("lib", "direct_dependency"),
                ),
            ),
            _SourceMock(
                filename=_path("collateral.vhd"),
                library="another_library",
                design_units=[{"name": "collateral_package", "type": "package"}],
                dependencies=(("work", "foo"), ("lib", "bar")),
            ),
        }

        self.database._configFromSources(sources, TEST_TEMP_PATH)

        # Check libraries before actually testing
        self.assertCountEqual(
            self.database.test_getDependenciesUnits(_Path("file_0.vhd")),
            {("some_library", "relative_dependency"), ("lib", "direct_dependency")},
        )
        self.assertCountEqual(
            self.database.test_getDependenciesUnits(_Path("collateral.vhd")),
            {("another_library", "foo"), ("lib", "bar")},
        )

        # Update the library of a path
        self.database._updatePathLibrary(
            _Path("file_0.vhd"), Identifier("file_0_lib", False)
        )

        # Check that reflected only where intended
        self.assertCountEqual(
            self.database.test_getDependenciesUnits(_Path("file_0.vhd")),
            {("file_0_lib", "relative_dependency"), ("lib", "direct_dependency")},
        )
        self.assertCountEqual(
            self.database.test_getDependenciesUnits(_Path("collateral.vhd")),
            {("another_library", "foo"), ("lib", "bar")},
        )

    def test_infers_uses_most_common_library_if_needed(self):
        # type: (...) -> Any
        # Given a design unit used in multiple ways, use the most common one
        self.database._configFromSources(
            {
                _SourceMock(
                    filename=_path("file_0.vhd"),
                    library=None,
                    design_units=[{"name": "package_0", "type": "package"}],
                    dependencies=(("lib_a", "some_dependency"),),
                ),
                _SourceMock(
                    filename=_path("file_1.vhd"),
                    library=None,
                    design_units=[{"name": "package_1", "type": "package"}],
                    dependencies=(("lib_b", "some_dependency"),),
                ),
                _SourceMock(
                    filename=_path("file_2.vhd"),
                    library=None,
                    design_units=[{"name": "package_2", "type": "package"}],
                    dependencies=(("lib_c", "some_dependency"),),
                ),
                _SourceMock(
                    filename=_path("file_3.vhd"),
                    library=None,
                    design_units=[{"name": "package_2", "type": "package"}],
                    dependencies=(("lib_a", "some_dependency"),),
                ),
                _SourceMock(
                    filename=_path("some_dependency.vhd"),
                    library=None,
                    design_units=[{"name": "some_dependency", "type": "package"}],
                ),
            },
            TEST_TEMP_PATH,
        )

        self.assertEqual(
            self.database.getLibrary(_Path("some_dependency.vhd")),
            Identifier("lib_a", False),
        )

        # Change one of the sources to use a different library to force the
        # most common one to change
        _SourceMock(
            filename=_path("file_0.vhd"),
            library=None,
            design_units=[{"name": "package_1", "type": "package"}],
            dependencies=(("lib_b", "some_dependency"),),
        )

        self.database.refresh()

        time.sleep(0.1)

        self.assertEqual(
            self.database.getLibrary(_Path("some_dependency.vhd")),
            Identifier("lib_b", False),
        )

    def test_json_encoding_and_decoding(self):
        database = _Database()

        database._configFromSources(
            {
                _SourceMock(
                    filename=_path("file_0.vhd"),
                    library="some_library",
                    design_units=[{"name": "some_package", "type": "package"}],
                    dependencies=(
                        ("work", "relative_dependency"),
                        ("lib", "direct_dependency"),
                    ),
                ),
                _SourceMock(
                    filename=_path("collateral.vhd"),
                    library="another_library",
                    design_units=[{"name": "collateral_package", "type": "package"}],
                    dependencies=(("work", "foo"), ("lib", "bar")),
                ),
            },
            TEST_TEMP_PATH,
        )

        state = json.dumps(database, cls=StateEncoder, indent=True)

        _logger.info("database in json:\n%s", state)

        recovered = json.loads(state, object_hook=jsonObjectHook)

        self.assertCountEqual(database.design_units, recovered.design_units)
        self.assertCountEqual(database._paths, recovered._paths)
        self.assertDictEqual(database._parse_timestamp, recovered._parse_timestamp)
        self.assertDictEqual(database._library_map, recovered._library_map)
        self.assertCountEqual(
            database._inferred_libraries, recovered._inferred_libraries
        )
        self.assertDictEqual(database._flags_map, recovered._flags_map)
        self.assertDictEqual(database._dependencies_map, recovered._dependencies_map)

    def test_removing_a_path_that_was_added(self):
        self.database._configFromSources(
            {
                _SourceMock(
                    filename=_path("file_0.vhd"),
                    library="some_library",
                    design_units=[{"name": "some_package", "type": "package"}],
                    dependencies=(
                        ("work", "relative_dependency"),
                        ("lib", "direct_dependency"),
                    ),
                ),
                _SourceMock(
                    filename=_path("collateral.vhd"),
                    library="another_library",
                    design_units=[{"name": "collateral_package", "type": "package"}],
                    dependencies=(("work", "foo"), ("lib", "bar")),
                ),
            },
            TEST_TEMP_PATH,
        )

        file_0 = _Path("file_0.vhd")
        collateral = _Path("collateral.vhd")
        # Make sure there's design units detected for all files
        self.assertTrue(list(self.database.getDesignUnitsByPath(file_0)))
        self.assertTrue(list(self.database.getDesignUnitsByPath(collateral)))
        self.assertEqual(str(self.database.getLibrary(file_0)), "some_library")
        self.assertEqual(str(self.database.getLibrary(collateral)), "another_library")

        self.database.removeSource(file_0)

        # file_0.vhd units will still be found (parsing does not depent on the
        # source being added or not)
        self.assertTrue(list(self.database.getDesignUnitsByPath(file_0)))
        # collateral.vhd units should continue to be found
        self.assertTrue(list(self.database.getDesignUnitsByPath(collateral)))

        # file_0.vhd is not on the project anymore, so library should reflect
        # that
        self.assertEqual(str(self.database.getLibrary(file_0)), "not_in_project")
        # collateral.vhd units should continue to be found
        self.assertEqual(str(self.database.getLibrary(collateral)), "another_library")

    def test_removing_an_existing_path_that_was_not_added(self):
        self.database._configFromSources(
            {
                _SourceMock(
                    filename=_path("collateral.vhd"),
                    library="another_library",
                    design_units=[{"name": "collateral_package", "type": "package"}],
                    dependencies=(("work", "foo"), ("lib", "bar")),
                )
            },
            TEST_TEMP_PATH,
        )

        file_0 = _Path("file_0.vhd")
        collateral = _Path("collateral.vhd")
        # Same as previous test, but file_0.vhd is not on the project
        self.assertTrue(list(self.database.getDesignUnitsByPath(file_0)))
        self.assertTrue(list(self.database.getDesignUnitsByPath(collateral)))
        # In this case, the library will change to the default library used for
        # files that weren't added
        self.assertEqual(str(self.database.getLibrary(file_0)), "not_in_project")
        self.assertEqual(str(self.database.getLibrary(collateral)), "another_library")

        self.database.removeSource(file_0)

        # file_0.vhd units will still be found (parsing does not depent on the
        # source being added or not)
        self.assertTrue(list(self.database.getDesignUnitsByPath(file_0)))
        # collateral.vhd units should continue to be found
        self.assertTrue(list(self.database.getDesignUnitsByPath(collateral)))

        # file_0.vhd is not on the project anymore, so library should reflect
        # that
        self.assertEqual(str(self.database.getLibrary(file_0)), "not_in_project")
        # collateral.vhd units should continue to be found
        self.assertEqual(str(self.database.getLibrary(collateral)), "another_library")

    def test_removing_a_non_existing_path_that_was_not_added(self):
        self.database._configFromSources(
            {
                _SourceMock(
                    filename=_path("collateral.vhd"),
                    library="another_library",
                    design_units=[{"name": "collateral_package", "type": "package"}],
                    dependencies=(("work", "foo"), ("lib", "bar")),
                )
            },
            TEST_TEMP_PATH,
        )

        path = Path("/some/path.vhd")
        self.assertFalse(p.exists(str(path)))

        # Try removing before any operation fills any table inside the
        # database. Nothing should change before and after
        before = self.database.__jsonEncode__()
        self.database.removeSource(path)
        self.assertDictEqual(self.database.__jsonEncode__(), before)

        collateral = _Path("collateral.vhd")

        # Same as previous test, but path.vhd is not on the project
        self.assertFalse(list(self.database.getDesignUnitsByPath(path)))
        self.assertTrue(list(self.database.getDesignUnitsByPath(collateral)))
        # In this case, the library will change to the default library used for
        # files that weren't added
        self.assertEqual(str(self.database.getLibrary(path)), "not_in_project")
        self.assertEqual(str(self.database.getLibrary(collateral)), "another_library")

        self.database.removeSource(path)

        # path.vhd units shouldn't be found anymore
        self.assertFalse(list(self.database.getDesignUnitsByPath(path)))
        # collateral.vhd units should continue to be found
        self.assertTrue(list(self.database.getDesignUnitsByPath(collateral)))

        # path.vhd is not on the project anymore
        self.assertEqual(str(self.database.getLibrary(path)), "not_in_project")
        # collateral.vhd units should continue to be found
        self.assertEqual(str(self.database.getLibrary(collateral)), "another_library")

        # Also test that a diagnostic indicating the path hasn't been added has
        # been created
        self.assertCountEqual(
            self.database.getDiagnosticsForPath(path), [PathNotInProjectFile(path)]
        )

    def test_temporary_paths_dont_generate_path_not_in_project(self):
        # type: (...) -> Any
        path = _path("foo.vhd")

        self.database._clearLruCaches()

        # Temporary paths should NOT generate PathNotInProjectFile diagnostic
        with patch.object(self.database, "_addDiagnostic") as meth:
            self.database.getLibrary(TemporaryPath(path))
            meth.assert_not_called()

        self.database._clearLruCaches()

        # Regular paths should generate PathNotInProjectFile diagnostic
        with patch.object(self.database, "_addDiagnostic") as meth:
            self.database.getLibrary(Path(path))
            meth.assert_called_once_with(PathNotInProjectFile(Path(path)))


class TestDirectDependencies(TestCase):
    def setUp(self):
        # type: (...) -> Any
        _logger.info("Setting up %s", self)
        self.database = _Database()

        sources = {
            _SourceMock(
                filename=_path("entity_a.vhd"),
                library="lib",
                design_units=[{"name": "entity_a", "type": "entity"}],
                dependencies=(
                    ("ieee", "numeric_std.all"),
                    ("common_dep",),
                    ("direct_dep_a",),
                    ("direct_dep_b",),
                ),
            ),
            _SourceMock(
                filename=_path("direct_dep_a.vhd"),
                library="lib",
                design_units=[
                    {"name": "direct_dep_a", "type": "entity"},
                    {"name": "side_effect_entity", "type": "entity"},
                ],
                dependencies=(("common_dep",), ("indirect_dep",)),
            ),
            _SourceMock(
                filename=_path("direct_dep_b.vhd"),
                library="lib",
                design_units=[{"name": "direct_dep_b", "type": "entity"}],
                dependencies=[("work", "common_dep")],
            ),
            _SourceMock(
                filename=_path("indirect_dep.vhd"),
                library="lib",
                design_units=[
                    {"name": "indirect_dep", "type": "package"},
                    {"name": "indirect_dep", "type": "package body"},
                    {"name": "side_effect_package", "type": "package"},
                ],
                dependencies=[("work", "common_dep")],
            ),
            _SourceMock(
                filename=_path("common_dep.vhd"),
                library="lib",
                design_units=[
                    {"name": "common_dep", "type": "package"},
                    {"name": "common_dep", "type": "package body"},
                ],
            ),
            _SourceMock(
                filename=_path("not_a_dependency.vhd"),
                library="lib",
                design_units=[{"name": "not_a_dependency", "type": "package"}],
                dependencies=(("common_dep",), ("indirect_dep",)),
            ),
        }

        self.database._configFromSources(sources, TEST_TEMP_PATH)

    def tearDown(self):
        # type: (...) -> Any
        _logger.info("Tearing down %s", self)
        self.database.test_reportCacheInfo()
        del self.database

    def test_get_correct_dependencies_of_entity_a(self):
        # type: (...) -> Any
        deps = list(self.database.test_getDependenciesUnits(_Path("entity_a.vhd")))

        # Indirect dependencies should always come first
        self.assertCountEqual(
            deps,
            {
                ("ieee", "numeric_std"),
                ("lib", "common_dep"),
                ("lib", "indirect_dep"),
                ("lib", "direct_dep_a"),
                ("lib", "direct_dep_b"),
            },
        )

    def test_get_correct_build_sequency_of_entity_a(self):
        # type: (...) -> Any
        sequence = list(self.database.test_getBuildSequence(_Path("entity_a.vhd")))

        # entity_a
        # '- common_dep
        # '- direct_dep_a
        #    '- common_dep
        #    '- indirect_dep
        # '- direct_dep_b
        #    '- common_dep

        common_dep = (Identifier("lib"), _Path("common_dep.vhd"))
        indirect_dep = (Identifier("lib"), _Path("indirect_dep.vhd"))
        direct_dep_a = (Identifier("lib"), _Path("direct_dep_a.vhd"))
        direct_dep_b = (Identifier("lib"), _Path("direct_dep_a.vhd"))

        # First item must be common dep
        self.assertEqual(sequence[0], common_dep)

        # indirect_dep must come before direct_dep_a
        self.assertIn(indirect_dep, sequence)
        self.assertIn(direct_dep_a, sequence)
        self.assertTrue(sequence.index(indirect_dep) < sequence.index(direct_dep_a))

        # If conditions above are respected, direct_dep_b can be anywhere
        self.assertIn(direct_dep_b, sequence)

    def test_get_correct_dependencies_of_indirect_dep(self):
        # type: (...) -> Any
        self.assertCountEqual(
            self.database.test_getDependenciesUnits(_Path("indirect_dep.vhd")),
            {("lib", "common_dep")},
        )


class TestDirectCircularDependencies(TestCase):
    def setUp(self):
        # type: (...) -> Any
        _logger.info("Setting up %s", self)
        self.database = _Database()

        sources = {
            _SourceMock(
                filename=_path("unit_a.vhd"),
                library="lib",
                design_units=[{"name": "unit_a", "type": "entity"}],
                dependencies=[("work", "unit_b")],
            ),
            _SourceMock(
                filename=_path("unit_b.vhd"),
                library="lib",
                design_units=[{"name": "unit_b", "type": "entity"}],
                dependencies=[("work", "unit_a")],
            ),
        }

        self.database._configFromSources(sources, TEST_TEMP_PATH)

    def tearDown(self):
        # type: (...) -> Any
        _logger.info("Tearing down %s", self)
        self.database.test_reportCacheInfo()
        del self.database

    def test_should_handle_both_sides(self):
        # type: (...) -> Any
        self.assertCountEqual(
            self.database.test_getDependenciesUnits(_Path("unit_a.vhd")),
            (("lib", "unit_b"),),
        )

        self.assertCountEqual(
            self.database.test_getDependenciesUnits(_Path("unit_b.vhd")),
            (("lib", "unit_a"),),
        )


class TestMultilevelCircularDependencies(TestCase):
    def setUp(self):
        # type: (...) -> Any
        _logger.info("Setting up %s", self)
        self.database = _Database()

        sources = {
            _SourceMock(
                filename=_path("unit_a.vhd"),
                library="work",
                design_units=[{"name": "unit_a", "type": "entity"}],
                dependencies=[("work", "unit_b")],
            ),
            _SourceMock(
                filename=_path("unit_b.vhd"),
                library="work",
                design_units=[{"name": "unit_b", "type": "entity"}],
                dependencies=[("work", "unit_c")],
            ),
            _SourceMock(
                filename=_path("unit_c.vhd"),
                library="work",
                design_units=[{"name": "unit_c", "type": "entity"}],
                dependencies=[("work", "unit_d")],
            ),
            _SourceMock(
                filename=_path("unit_d.vhd"),
                library="work",
                design_units=[{"name": "unit_d", "type": "entity"}],
                dependencies=[("work", "unit_a")],
            ),
            _SourceMock(
                filename=_path("not_a_dependency.vhd"),
                library="work",
                design_units=[{"name": "not_a_dependency", "type": "package"}],
                dependencies=[("indirect_dep",), ("common_dep",)],
            ),
        }

        self.database._configFromSources(sources, TEST_TEMP_PATH)

    def tearDown(self):
        # type: (...) -> Any
        _logger.info("Tearing down %s", self)
        self.database.test_reportCacheInfo()
        del self.database

    def test_report_all_but_the_source_in_question(self):
        # type: (...) -> Any
        self.assertCountEqual(
            self.database.test_getDependenciesUnits(_Path("unit_a.vhd")),
            {("work", "unit_b"), ("work", "unit_c"), ("work", "unit_d")},
        )

        self.assertCountEqual(
            self.database.test_getDependenciesUnits(_Path("unit_b.vhd")),
            {("work", "unit_a"), ("work", "unit_c"), ("work", "unit_d")},
        )

        self.assertCountEqual(
            self.database.test_getDependenciesUnits(_Path("unit_c.vhd")),
            {("work", "unit_a"), ("work", "unit_b"), ("work", "unit_d")},
        )

        self.assertCountEqual(
            self.database.test_getDependenciesUnits(_Path("unit_d.vhd")),
            {("work", "unit_a"), ("work", "unit_b"), ("work", "unit_c")},
        )


class TestIndirectLibraryInference(TestCase):
    def setUp(self):
        # type: (...) -> Any
        _logger.info("Setting up %s", self)
        self.database = _Database()

        sources = {
            _SourceMock(
                filename=_path("target_pkg.vhd"),
                library=None,
                design_units=[{"name": "target_pkg", "type": "package"}],
            ),
            _SourceMock(
                filename=_path("no_lib_but_use_it_directly.vhd"),
                library=None,
                design_units=[
                    {"name": "no_lib_but_use_it_directly", "type": "package"}
                ],
                dependencies=[("find_me", "target_pkg")],
            ),
            _SourceMock(
                filename=_path("with_lib_but_use_it_directly.vhd"),
                library="find_me",
                design_units=[
                    {"name": "with_lib_but_use_it_directly", "type": "package"}
                ],
                dependencies=[("work", "target_pkg")],
            ),
        }

        self.database._configFromSources(sources, TEST_TEMP_PATH)

    def tearDown(self):
        # type: (...) -> Any
        _logger.info("Tearing down %s", self)
        self.database.test_reportCacheInfo()
        del self.database

    def test_infer_library_when_using_directly(self):
        # type: (...) -> Any
        sequence = tuple(
            self.database.test_getBuildSequence(_Path("no_lib_but_use_it_directly.vhd"))
        )

        self.assertEqual(sequence, ((Identifier("find_me"), _Path("target_pkg.vhd")),))

    def test_infer_library_from_path(self):
        # type: (...) -> Any
        sequence = tuple(
            self.database.test_getBuildSequence(
                _Path("with_lib_but_use_it_directly.vhd")
            )
        )

        self.assertEqual(sequence, ((Identifier("find_me"), _Path("target_pkg.vhd")),))


class TestUnitsDefinedInMultipleSources(TestCase):
    def setUp(self):
        # type: (...) -> Any
        _logger.info("Setting up %s", self)
        self.database = _Database()

        sources = {
            _SourceMock(
                filename=_path("no_lib_target.vhd"),
                library=None,
                design_units=[{"name": "no_lib_target", "type": "entity"}],
                dependencies=[
                    ("work", "no_lib_package"),
                    ("work", "dependency"),
                    ("work", "no_lib_package"),
                ],
            ),
            _SourceMock(
                filename=_path("dependency.vhd"),
                library=None,
                design_units=[{"name": "dependency", "type": "package"}],
            ),
            _SourceMock(
                filename=_path("no_lib_package_1.vhd"),
                library=None,
                design_units=[{"name": "no_lib_package", "type": "package"}],
            ),
            _SourceMock(
                filename=_path("no_lib_package_2.vhd"),
                library=None,
                design_units=[{"name": "no_lib_package", "type": "package"}],
            ),
            _SourceMock(
                filename=_path("collateral.vhd"),
                library=None,
                design_units=[{"name": "collateral", "type": "package"}],
            ),
        }

        self.database._configFromSources(sources, TEST_TEMP_PATH)

    def tearDown(self):
        # type: (...) -> Any
        _logger.info("Tearing down %s", self)
        self.database.test_reportCacheInfo()
        del self.database

    def test_build_sequence_units_are_unique(self):
        # type: (...) -> Any
        # If a design unit is defined in multiple places, we should not include
        # all of them

        # The exact one picked is not fixed (but it has to be constant
        # throughout a session to avoid bouncing)
        self.assertIn(
            set(self.database.test_getBuildSequence(_Path("no_lib_target.vhd"))),
            (
                {
                    (Identifier("library"), _Path("dependency.vhd")),
                    (Identifier("library"), _Path("no_lib_package_1.vhd")),
                },
                {
                    (Identifier("library"), _Path("dependency.vhd")),
                    (Identifier("library"), _Path("no_lib_package_2.vhd")),
                },
            ),
        )

    def test_non_unique_units_are_reported(self):
        # type: (...) -> Any

        # Design units defined in multiple places should trigger
        # DependencyNotUnique in the path's diagnostics

        # Need to get the build sequence before this diag is populated
        list(self.database.test_getBuildSequence(_Path("no_lib_target.vhd")))

        logIterable("all diags", self.database._diags.items(), _logger.info)

        self.assertCountEqual(
            {
                DependencyNotUnique(
                    filename=_Path("no_lib_target.vhd"),
                    design_unit=DependencySpec(
                        owner=_Path("no_lib_target.vhd"),
                        name=Identifier("no_lib_package", False),
                    ),
                    choices={
                        _Path("no_lib_package_1.vhd"),
                        _Path("no_lib_package_2.vhd"),
                    },
                    line_number=2,
                    column_number=5,
                ),
                DependencyNotUnique(
                    filename=_Path("no_lib_target.vhd"),
                    design_unit=DependencySpec(
                        owner=_Path("no_lib_target.vhd"),
                        name=Identifier("no_lib_package", False),
                    ),
                    choices={
                        _Path("no_lib_package_1.vhd"),
                        _Path("no_lib_package_2.vhd"),
                    },
                    line_number=4,
                    column_number=5,
                ),
            },
            self.database.getDiagnosticsForPath(_Path("no_lib_target.vhd")),
        )
        self.assertNotEqual(
            list(
                self.database.getPathsDefining(
                    name=Identifier("no_lib_package"), library=None
                )
            ),
            [],
        )

    def test_temporary_paths_dont_generate_diagnostics(self):
        # type: (...) -> Any

        # Create a copy of a source file with same contents but a different
        # name to mimic getting info from a dump (which is in itself a version
        # of an existing file)
        _SourceMock(
            filename=_path("no_lib_package_3.vhd"),
            library=None,
            design_units=[{"name": "no_lib_package", "type": "package"}],
        )
        # This file should be added so that the database is aware of its
        # existence.

        # Add this as a regular file to make sure the test will fail
        self.database.addSource(_Path("no_lib_package_3.vhd"), None)
        with self.assertRaises(AssertionError):
            self.test_non_unique_units_are_reported()
        self.database.removeSource(_Path("no_lib_package_3.vhd"))

        # Now add the same path using TemporaryPath, which should make the
        # previous test pass
        self.database.addSource(TemporaryPath(_path("no_lib_package_3.vhd")), None)

        # Test should run exactly the same as before. If we added using the
        # path.Path class, the test would have failed
        self.test_non_unique_units_are_reported()

    def test_temporary_paths_are_excluded(self):
        # type: (...) -> Any
        # This should add diagnostics
        with patch.object(self.database, "_addDiagnostic") as meth:
            name = Identifier("no_lib_package")
            choices = {_Path("no_lib_package_1.vhd"), _Path("no_lib_package_2.vhd")}
            self.database._reportDependencyNotUnique(
                name=name, library=None, choices=choices
            )

            meth.assert_called()

        # This should not
        with patch.object(self.database, "_addDiagnostic") as meth:
            name = Identifier("no_lib_package")
            choices = {
                _Path("no_lib_package_1.vhd"),
                TemporaryPath(_path("no_lib_package_2.vhd")),
            }
            self.database._reportDependencyNotUnique(
                name=name, library=None, choices=choices
            )

            meth.assert_not_called()



# This file is part of HDL Checker.
#
# Copyright (c) 2015 - 2019 suoto (Andre Souto)
#
# HDL Checker is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HDL Checker is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with HDL Checker.  If not, see <http://www.gnu.org/licenses/>.
"Base class that implements the base builder flow"

import logging
import os.path as p
from contextlib import contextmanager
from enum import Enum
from tempfile import mkdtemp
from typing import (  # pylint: disable=unused-import
    Any,
    Dict,
    Iterable,
    List,
    Tuple,
    Union,
)

from .builders.fallback import Fallback
from .builders.ghdl import GHDL
from .builders.msim import MSim
from .builders.xvhdl import XVHDL

from hdl_checker.parsers.elements.identifier import Identifier
from hdl_checker.path import Path  # pylint: disable=unused-import
from hdl_checker.types import BuildFlags, FileType
from hdl_checker.utils import removeDirIfExists

_logger = logging.getLogger(__name__)

AnyValidBuilder = Union[MSim, XVHDL, GHDL]
AnyBuilder = Union[AnyValidBuilder, Fallback]


class BuilderName(Enum):
    """
    Supported tools
    """

    msim = MSim.builder_name
    xvhdl = XVHDL.builder_name
    ghdl = GHDL.builder_name
    fallback = Fallback.builder_name


def getBuilderByName(name):
    "Returns the builder class given a string name"
    # Check if the builder selected is implemented and create the
    # builder attribute
    if name == "msim":
        builder = MSim
    elif name == "xvhdl":
        builder = XVHDL
    elif name == "ghdl":
        builder = GHDL
    else:
        builder = Fallback

    return builder


def getWorkingBuilders():
    """
    Returns a generator with the names of builders that are actually working
    """
    for builder_class in AVAILABLE_BUILDERS:
        if builder_class is Fallback:
            continue
        if builder_class.isAvailable():
            _logger.debug("Builder %s worked", builder_class.builder_name)
            yield builder_class


def foundVunit():  # type: () -> bool
    """
    Checks if our env has VUnit installed
    """
    try:
        import vunit  # type: ignore pylint: disable=unused-import

        return True
    except ImportError:  # pragma: no cover
        pass

    return False


_VUNIT_FLAGS = {
    BuilderName.msim: {"93": ("-93",), "2002": ("-2002",), "2008": ("-2008",)},
    BuilderName.ghdl: {
        "93": ("--std=93c",),
        "2002": ("--std=02",),
        "2008": ("--std=08",),
    },
}  # type: Dict[BuilderName, Dict[str, BuildFlags]]


def getVunitSources(builder):
    # type: (AnyValidBuilder) -> Iterable[Tuple[Path, str, BuildFlags]]
    "Foo bar"
    if not foundVunit():  # or self._builder_name == BuilderName.fallback:
        return

    import vunit  # pylint: disable=import-error

    logging.getLogger("vunit").setLevel(logging.ERROR)

    _logger.debug("VUnit installation found")

    sources = []  # type: List[Any]

    # Prefer VHDL VUnit
    if FileType.vhdl in builder.file_types:
        from vunit import VUnit  # pylint: disable=import-error

        sources += _getSourcesFromVUnitModule(VUnit)
        _logger.debug("Added VUnit VHDL files")

    if FileType.systemverilog in builder.file_types:
        from vunit.verilog import (  # type: ignore # pylint: disable=import-error
            VUnit,
        )

        _logger.debug("Builder supports Verilog, adding VUnit Verilog files")
        builder.addExternalLibrary(FileType.verilog, Identifier("vunit_lib", False))
        builder.addIncludePath(
            FileType.verilog, p.join(p.dirname(vunit.__file__), "verilog", "include")
        )
        sources += _getSourcesFromVUnitModule(VUnit)

    if not sources:
        _logger.info("Vunit found but no file types are supported by %s", builder)
        return

    for source in sources:
        path = p.abspath(source.name)
        library = source.library.name

        # Get extra flags for building VUnit sources
        try:
            flags = _VUNIT_FLAGS[BuilderName(builder.builder_name)][
                source.vhdl_standard
            ]
        except KeyError:
            flags = tuple()

        yield Path(path), library, flags


@contextmanager
def _makeTemporaryDir(*args, **kwargs):
    """
    Context manager that wraps tempfile.mkdtemp but deletes the directory
    afterwards
    """
    path = mkdtemp(*args, **kwargs)
    yield path
    removeDirIfExists(path)


def _getSourcesFromVUnitModule(vunit):
    """
    Creates a temporary VUnit project given a VUnit module and return a list of
    its files
    """
    with _makeTemporaryDir() as output_path:

        # Create a dummy VUnit project to get info on its sources
        vunit_project = vunit.from_argv(["--output-path", output_path])

        # OSVVM is always avilable
        vunit_project.add_osvvm()
        # Communication library and array utility library are only
        # available on VHDL 2008
        if vunit_project.vhdl_standard == "2008":
            vunit_project.add_com()
            vunit_project.add_array_util()

        return list(vunit_project.get_source_files())


__all__ = ["MSim", "XVHDL", "GHDL", "Fallback"]

AVAILABLE_BUILDERS = MSim, XVHDL, GHDL, Fallback
